# coding: utf-8
"""
    SecTools.CorMap.ScaledDataDialog.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import os
import logging
from bisect import bisect_right
import numpy as np
from scipy.optimize import minimize
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.patches import Rectangle
from matplotlib.widgets import SpanSelector
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
from SerialDataUtils import load_intensity_files
from SaferSpinbox import SaferSpinbox
from molass_legacy.KekLib.TkSupplements import BlinkingFrame
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from molass_legacy.KekLib.TkUtils import split_geometry
from MatrixData import simple_plot_3d
from molass_legacy._MOLASS.SerialSettings import get_setting
from DataUtils import get_in_folder
from .AngularUnit import angstrom_qv
from .CormapMaker import CormapMaker
from Theory.SolidSphere import get_boundary_params_simple

ALLOW_SPAN_CHANGE = False
DATA_VIESW_NAMES = ["2D View", "3D View"]
CORREL_TYPE_NAMES = ["numpy corrcoef", "ATSAS DATCMP"]
TRANSPOSE_NAMES = ["Frames", "Angular"]

def get_version_string():
    import platform
    return 'Scaled Cormap 0.0.1 (2021-12-21 python %s %s)' % (
                platform.python_version(), platform.architecture()[0])

class ScaledDataDialog(Dialog):
    def __init__(self, parent, demo=False, elution=False, atsas=None):
        self.logger = logging.getLogger(__name__)
        self.demo = demo
        self.elution = elution
        if atsas is None:
            from Env.EnvInfo import get_global_env_info
            env_info = get_global_env_info()
            atsas = env_info.atsas_is_available
        self.threed = False
        self.atsas = atsas
        self.rect = None
        self.colorbar = None
        self.cbar_ax = None
        self.cormap_drawn = False
        self.showing_transposed = True
        self.showing_datcmp = False
        self.datafiles = None
        self.popup_menu1 = None
        self.popup_menu3 = None
        self.datcmp_data = None
        Dialog.__init__(self, parent, get_version_string(), visible=False)

    def body(self, body_frame):
        frame = Tk.Frame(body_frame)
        frame.pack(padx=20, pady=10)

        input_frame = Tk.Frame(frame)
        input_frame.pack(pady=20)

        label = Tk.Label(input_frame, text="Input Folder: ")
        label.pack(side=Tk.LEFT)

        self.in_folder = Tk.StringVar()
        self.fe = FolderEntry(input_frame, textvariable=self.in_folder, width=80,
                                on_entry_cb=self.on_in_folder_entry)
        self.fe.pack(side=Tk.LEFT)

        cframe = Tk.Frame(frame)
        cframe.pack(padx=20)

        self.fig = fig = plt.figure(figsize=(18,6))
        ax1 = fig.add_subplot(131, projection="3d")
        ax2 = fig.add_subplot(132, projection="3d")
        ax3 = fig.add_subplot(133)
        self.axes = ax1, ax2, ax3

        self.fig.tight_layout()
        # self.fig.subplots_adjust(top=0.85, left=0.05, right=0.94, bottom=0.1, wspace=1.8)
        self.fig.subplots_adjust(top=0.85, left=0.05, right=0.94, bottom=0.1)

        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        tframe = Tk.Frame(frame)
        tframe.pack(padx=0, fill=Tk.X)      # padx=0: suppress changes from toolbar cursor display
        tframe_left = Tk.Frame(tframe)
        tframe_left.pack(side=Tk.LEFT, fill=Tk.X)
        tframe_right = Tk.Frame(tframe)
        tframe_right.pack(side=Tk.RIGHT)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe_left)
        self.toolbar.update()

        self.build_control_bar(tframe_right)

    def reset_cbar_ax(self):
        if self.cbar_ax is not None:
            self.cbar_ax.remove()
        self.cbar_ax = self.fig.add_axes([0.89, 0.0, 0.06, 1.0])
        self.cbar_ax.set_axis_off()

    def build_control_bar(self, frame):
        pass

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=50, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def show(self):
        if self.demo:
            self.after(500, self.perpare_temp_demo)
        self._show()

    def perpare_temp_demo(self):
        from molass_legacy._MOLASS.SerialSettings import get_setting
        in_folder = get_setting("in_folder")
        self.in_folder.set(in_folder)
        self.on_in_folder_entry()

    def on_in_folder_entry(self):
        self.update()
        in_folder = self.in_folder.get()
        try:
            data_array, datafiles = load_intensity_files(in_folder)
            self.datafiles = datafiles
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(None, "on_in_folder_entry: ")
            return

        self.set_data(data_array)

        in_folder_ = get_in_folder(in_folder.replace("\\", "/"))
        self.fig.suptitle("Correlation Analysis on %s" % in_folder_, fontsize=20)
        M_, qv_, aslice, eslice= self.compute_scaled_data()
        self.draw_input_data(M_, qv_, aslice, eslice)
        self.draw_scaled_data(M_, qv_, aslice, eslice)
        self.draw_cormap(M_, qv_, aslice, eslice)
        self.mpl_canvas.draw()

    def set_data(self, data_array):
        from molass_legacy.Elution.CurveUtils import get_xray_elution_vector
        from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
        self.qv = angstrom_qv(data_array[0,:,0])
        self.D = data_array[:,:,1].T
        self.E = data_array[:,:,2].T
        y, slice_ = get_xray_elution_vector(self.qv, data_array)
        self.xray_slice = slice_
        self.xray_index = (slice_.start + slice_.stop)//2
        self.ecurve = ElutionCurve(y)

    def get_range_info(self):
        from molass_legacy.Trimming.PreliminaryRg import PreliminaryRg
        from molass_legacy.Trimming.FlangeLimit import FlangeLimit
        from molass_legacy.Trimming.GuinierLimit import GuinierLimit

        ecurve = self.ecurve

        pinfo = ecurve.get_primary_peak_info()
        f, t = pinfo[0], pinfo[2]
        eslice = slice(f, t+1)

        f_limit = FlangeLimit(self.D, self.E, ecurve, self.qv).get_limit()
        pre_rg = PreliminaryRg(self.D, self.E, ecurve, self.qv, f_limit)
        acceptable_rg_consist = get_setting("acceptable_rg_consist")
        g_limit = GuinierLimit(self.D, self.ecurve, pre_rg, f_limit).get_limit(acceptable_rg_consist)

        b1, b2, k = get_boundary_params_simple(pre_rg.Rg)
        print("Rg=%.3g, b1=%.3g, b2=%.3g, k=%.3g" % (pre_rg.Rg, b1, b2, k))
        i = bisect_right(self.qv, b1)

        aslice = slice(g_limit, i)
        return aslice, eslice

    def compute_scaled_data(self):

        aslice, eslice = self.get_range_info()
        M = self.D[aslice,eslice]
        qv_ = self.qv[aslice]

        pinfo = self.ecurve.get_primary_peak_info()
        ptj = pinfo[1]
        pt_ridge_y = np.average(self.D[aslice,ptj-2:ptj+3], axis=1)

        scaled_list = []
        for k in range(M.shape[1]):
            y = M[:,k]
            def obj_func(p):
                return np.sum((y*p[0] - pt_ridge_y)**2)

            ret = minimize(obj_func, (1,))
            scaled_list.append(y*ret.x[0])

        M_ = np.array(scaled_list).T

        return M_, qv_, aslice, eslice

    def draw_input_data(self, M_, qv_, aslice, eslice):
        ax = self.axes[0]
        ax.cla()
        ax.set_title("Input Data", fontsize=16)
        simple_plot_3d(ax, self.D, x=self.qv)

        x = self.qv[aslice]
        for j in [eslice.start, eslice.stop-1]:
            y = np.ones(len(x))*j
            z = self.D[aslice,j]
            ax.plot(x, y, z, color="cyan")

        y = self.ecurve.x
        z = self.ecurve.y
        x = np.ones(len(y)) * self.qv[self.xray_index]
        ax.plot(x, y, z, color="orange")

    def draw_scaled_data(self, M_, qv_, aslice, eslice):

        frames = np.arange(eslice.start, eslice.stop)
        ax = self.axes[1]
        ax.cla()
        ax.set_title("Scaled Data", fontsize=16)
        simple_plot_3d(ax, M_, x=qv_, y=frames)

        j0 = eslice.start
        j1 = eslice.stop - 1
        x = qv_
        for j in [j0, j1]:
            y = np.ones(len(x))*j
            z = M_[:,j - j0]
            ax.plot(x, y, z, color="cyan")

        y = self.ecurve.x[eslice]
        i = self.xray_index - aslice.start
        if i < M_.shape[0]:
            z = M_[i,:]
            x = np.ones(len(y)) * self.qv[self.xray_index]
            ax.plot(x, y, z, color="orange")
        else:
            # as in 20161216/Backsub
            pass

    def draw_cormap(self, M_, qv_, aslice, eslice):
        from molass_legacy.ATSAS.DatCmp import run_datcmp_from_array

        E_ = self.E[aslice,eslice]
        datcmp_data = run_datcmp_from_array(qv_, M_, E_, P_threshold=0.05)

        ax = self.axes[2]
        ax.cla()
        self.reset_cbar_ax()

        ax.set_title("CorMap of Scaled Data", fontsize=16)

        frames = np.arange(eslice.start, eslice.stop)
        cm = CormapMaker(datcmp_data, qv_, frames, "Frame No.", transpose=True, from_datcmp=True)
        cm.draw(ax, self.cbar_ax)

    def save_the_figure(self, file):
        self.fig.savefig(file)



"""
    Radial Distribution Function
    https://physicspython.wordpress.com/2019/07/31/radial-distribution-function/
"""
import scipy as sp
from matplotlib import pyplot as plt
 
def volume(z, r, z_bot, z_top):
    """ volume of a sphere of radius r located at height z """
    volume = 4.0 / 3.0 * sp.pi * r**3
     
    """ cut off a spherical cap if the sphere extends below z_bot """
    if z - r < z_bot:
        h = z_bot - (z - r)
        volume -= sp.pi * h**2 * (r - h / 3.0)
 
    """ cut off a spherical cap if the sphere extends above z_top """
    if z + r > z_top:
        h = z + r - z_top
        volume -= sp.pi * h**2 * (r - h / 3.0)
     
    return volume
 
def distance(a, b):
    """ get displacement in each coordinate and wrap w.r.t. lattice parameter """
    dx = abs(a[0] - b[0])
    x = min(dx, abs(A - dx))
     
    dy = abs(a[1] - b[1])
    y = min(dy, abs(B - dy))
     
    dz = abs(a[2] - b[2])
    z = min(dz, abs(C - dz))
     
    return sp.sqrt(x**2 + y**2 + z**2)
 
class Trajectory:
    def __init__(self, filename, skip, z_bot_interface, z_top_interface, interface_offset=4.0, resolution=200):
        """ 
        filename         : path to the trajectory file 
        skip             : number of snapshots to be skipped between two configurations that are evaluated
                           (for example, if trajectory is 9000 steps long, and skip = 10, every tenth step
                           is evaluated, 900 steps in total; use skip = 1 to take every step of the MD)
        z_bot_interface  : average vertical coordinate for interface below water layer in Angstrom
        z_top_interface  : average vertical coordinate for interface above water layer in Angstrom
        interface_offset : distance between interface and region of water with bulk-like properties
        resolution       : number of points in the final radial distribution function """
         
        with open(filename, 'r') as f:
            data = f.readlines()
         
        self.z_top = z_top_interface - interface_offset
        self.z_bot = z_bot_interface + interface_offset
         
        self.n_atoms = int(data[0].split()[0])
        self.n_steps_total = int(len(data) / (self.n_atoms + 2))
         
        self.atom_list = [line.split()[0] for line in data[2 : self.n_atoms + 2]]
         
        self.skip = skip
        self.n_steps = self.n_steps_total // self.skip
         
        self.coordinates = sp.zeros((self.n_steps, self.n_atoms, 3))
         
        for step in range(self.n_steps):
            coords = sp.zeros((self.n_atoms, 3))
             
            i = step * self.skip * (self.n_atoms + 2)
            for j, line in enumerate(data[i + 2 : i + self.n_atoms + 2]):
                coords[j, :] = [float(value) for value in line.split()[1:]]
             
            self.coordinates[step] = coords
         
        self.resolution = resolution
        self.compute_volume_per_h2o()
     
    def compute_volume_per_h2o(self):
        """ calculates the average volume per water molecule from the subset of configurations """
        """ sum up all oxygens belonging to water in the bulk region """
        n_h2o = 0
        for step in range(self.n_steps):
            for i, atom in enumerate(self.coordinates[step]):
                if self.atom_list[i] == "O" and self.z_bot < atom[2] < self.z_top:
                    n_h2o += 1
         
        """ devide the volume of the bulk region by the average number of molecules per step """
        self.volume_per_h2o = A * B * (self.z_top - self.z_bot) * self.n_steps / n_h2o
     
    def compute_radial_distribution(self):
        """ no reason to go above half of the smallest lattice parameter as mirror images start 
        to be double-counted """
        r_cutoff = min(A, B) / 2.0
        dr = r_cutoff / self.resolution
        volumes = sp.zeros(self.resolution)
         
        self.radii = sp.linspace(0.0, self.resolution * dr, self.resolution)
        self.g_of_r = sp.zeros(self.resolution)
         
        for step in range(self.n_steps):
            print('{:4d} : {:4d}'.format(step, self.n_steps))
             
            """ isolate all liquid water molecules based on the position of the oxygen atoms """
            data_oxygen = []
            for i, atom in enumerate(self.coordinates[step]):
                if self.atom_list[i] == "O":
                    if self.z_bot < atom[2] < self.z_top:
                        data_oxygen.append(atom)
            data_oxygen = sp.array(data_oxygen)
             
            """ loop over each pair of H2O molecules, calculate their distance, build a histogram 
            each pair is evaluated as two contributions to the distribution function """
            for i, oxygen1 in enumerate(data_oxygen):
                for j in range(self.resolution):
                    r1 = j * dr
                    r2 = r1 + dr
                    v1 = volume(oxygen1[2], r1, self.z_bot, self.z_top)
                    v2 = volume(oxygen1[2], r2, self.z_bot, self.z_top)
                    volumes[j] += v2 - v1
                 
                for oxygen2 in data_oxygen[i:]:
                    dist = distance(oxygen1, oxygen2)
                    index = int(dist / dr)
                    if 0 < index < self.resolution:
                        self.g_of_r[index] += 2.0
         
        """ normalize by the volume of the spherical shell corresponding to each radius """
        for i, value in enumerate(self.g_of_r):
            self.g_of_r[i] = value * self.volume_per_h2o / volumes[i]
         
    def plot(self, filename=""):
        """ plots the radial distribution function
        if filename is specified, prints it to file as a pdf """
         
        if not self.g_of_r.any():
            print('compute the radial distribution function first\n')
            return
         
        plt.xlabel('r (Å)')
        plt.ylabel('g$_{O-O}$(r)')
        plt.plot(self.radii, self.g_of_r)
         
        if filename:
            plt.savefig('radial_distribution_function.pdf', dpi=300, bbox='tight', format='pdf')
         
        plt.show()
 
 
A = 12.991
B = 11.832
C = 30.577
bottom_interface = 23.0
top_interface = 40.0
 
TiO2_H2O = Trajectory('TiO2_H2O-pos-1.xyz', 10, 
                      bottom_interface, 
                      top_interface)
TiO2_H2O.compute_radial_distribution()
TiO2_H2O.plot()



# coding: utf-8
"""
    ReAtsas.Almerge.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""

import numpy as np
import molass_legacy.KekLib.DebugPlot as dplt
from SerialAtsasTools import AlmergeExecutor

class Almerge(AlmergeExecutor):
    def __init__(self):
        AlmergeExecutor.__init__(self)



# coding: utf-8
"""
    KL_Divergence.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.stats import entropy, mode
from scipy.optimize import minimize

def synchronize(pd1, pd2, distance=entropy):
    if pd1.N >= pd2.N:
        pd1_, pd2_ = pd1, pd2
    else:
        pd1_, pd2_ = pd2, pd1

    j = pd2_.x

    def obj_function(p):
        A, B = p
        print(A,B)
        i = A*j + B
        return distance(pd1_.pdf(i), pd2_.ny)    # return KL divergence

    A_init = pd1_.N/pd2_.N
    m1 = pd1_.mode
    m2 = pd2_.mode
    B_init = m1 - m2*A_init
    ratio = abs(B_init)/pd1.N
    if ratio > 0.03:
        print('using means', ratio)
        m1 = pd1_.M1
        m2 = pd2_.M1
        B_init = m1 - m2*A_init
    else:
        print('using modes', ratio)
    result = minimize(obj_function, [A_init, B_init], args=())

    A, B = result.x
    if pd1.N < pd2.N:
        A, B = 1/A, -B/A

    return A, B



"""
    SMC.CloudPickleTest.py
"""
from time import sleep
import multiprocessing
import cloudpickle
from concurrent.futures import ProcessPoolExecutor, wait
from pymc.util import CustomProgress
from rich.progress import (
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)

class Problematic:
    def __init__(self):
        pass
    def _method(self):
        pass

def exec_func(
    chain,
    progress_dict,
    task_id,
):
    sleep(3)
    problem = Problematic()
    results = (
        chain,
        problem,
    )
    return cloudpickle.dumps(results)

def run_processes():
    with CustomProgress(
        TextColumn("{task.description}"),
        SpinnerColumn(),
        TimeRemainingColumn(),
        TextColumn("/"),
        TimeElapsedColumn(),
        TextColumn("{task.fields[status]}"),
        disable=False,
    ) as progress:
        kwargs = {'a':1, 'b':2}
        workers = 4
        chains = 2
        futures = []  # keep track of the jobs
        with multiprocessing.Manager() as manager:
            # this is the key - we share some state between our
            # main process and our worker functions
            _progress = manager.dict()

            with ProcessPoolExecutor(max_workers=workers) as executor:
                for c in range(chains):  # iterate over the jobs we need to run
                    # set visible false so we don't have a lot of bars all at once:
                    task_id = progress.add_task(f"Chain {c}", status="Stage: 0 Beta: 0")
                    futures.append(
                        executor.submit(
                            exec_func,
                            c,
                            _progress,
                            task_id,
                        )
                    )

                # monitor the progress:
                done = []
                remaining = futures
                while len(remaining) > 0:
                    finished, remaining = wait(remaining, timeout=0.1)
                    done.extend(finished)
                    for task_id, update_data in _progress.items():
                        stage = update_data["stage"]
                        beta = update_data["beta"]
                        # update the progress bar for this task:
                        progress.update(
                            status=f"Stage: {stage} Beta: {beta:.3f}",
                            task_id=task_id,
                            refresh=True,
                        )

    return tuple(cloudpickle.loads(r.result()) for r in done)

if __name__ == '__main__':
    run_processes()


# coding: utf-8
"""
    ConcOptsDialog.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.TkUtils import rational_geometry
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable, BlinkingFrame
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting

class ConcOptsDialog( Dialog ):
    def __init__(self, parent, dialog, pranges):
        # print('pranges=', pranges)
        self.dialog = dialog
        self.pranges = pranges
        self.sd = dialog.sd
        self.xr_j0 = dialog.sd.xr_j0
        self.range_id_list = []
        self.range_id_dict = {}
        for k, r in enumerate(pranges):
            ft_list = r.get_fromto_list()
            print([k], ft_list)
            for m, ft in enumerate(ft_list):
                ft_ = [ self.xr_j0 + j for j in ft ]
                rande_id = '%d-%d (%s)' % (k+1, m+1, str(ft_))
                self.range_id_dict[rande_id] = ft_
                self.range_id_list.append(rande_id)
        Dialog.__init__( self, parent, "Concentration Options", visible=False, geometry_cb=self.adjust_geometry )

    def adjust_geometry(self):
        rational_geometry(self, self.dialog, 0.5, 0.5)

    def show(self):
        self._show()

    def body( self, body_frame ):
        tk_set_icon_portable( self )

        iframe = Tk.Frame(body_frame)
        iframe.pack(padx=20, pady=10)

        grid_row = 0

        label = Tk.Label(iframe, text="Scattering Curve Inspection View: ")
        label.grid(row=grid_row, column=0, sticky=Tk.W)

        label = Tk.Label(iframe, text="Range ID Selection")
        label.grid(row=grid_row, column=1)

        self.range_id = Tk.StringVar()
        self.range_id.set(self.range_id_list[0])
        range_id_box = ttk.Combobox( iframe, textvariable=self.range_id, width=16, justify=Tk.CENTER )
        range_id_box[ 'values' ] = self.range_id_list
        range_id_box.grid(row=grid_row, column=2)

        grid_row += 1
        plot_button = Tk.Button(iframe, text="Show", command=self.plot_normalized_curves)
        plot_button.grid(row=grid_row, column=1, columnspan=2, pady=10)

        grid_row += 1
        space = Tk.Label(iframe)
        space.grid(row=grid_row, column=0)

        grid_row += 1
        label = Tk.Label(iframe, text="Concentration Type: ")
        label.grid(row=grid_row, column=0, sticky=Tk.W)

        self.conc_curve_type = Tk.IntVar()
        self.conc_curve_type.set(get_setting('conc_curve_type'))

        for k, cname in enumerate([ 'Synchronized UV Elution Curve', 'Scaled Xray Elution Curve', ]):
            rb = Tk.Radiobutton( iframe, text=cname,
                        variable=self.conc_curve_type, value=k,
                        )
            rb.grid( row=grid_row+k, column=1, sticky=Tk.W )

        grid_row += 2
        space = Tk.Label(iframe)
        space.grid(row=grid_row, column=0)

        grid_row += 1
        label = Tk.Label(iframe, text="Weight Matrix Type: ")
        label.grid(row=grid_row, column=0, sticky=Tk.W)

        self.weight_matrix_type = Tk.IntVar()
        self.weight_matrix_type.set(get_setting('weight_matrix_type'))

        for k, cname in enumerate([ 'Uniform', 'Concentration Reciprocal', ]):
            rb = Tk.Radiobutton( iframe, text=cname,
                        variable=self.weight_matrix_type, value=k,
                        )
            rb.grid( row=grid_row+k, column=1, sticky=Tk.W )

    def apply(self):
        set_setting('conc_curve_type', self.conc_curve_type.get())
        set_setting('weight_matrix_type', self.weight_matrix_type.get())

    def plot_normalized_curves(self):
        from ScatteringCurveViewer import ScatteringCurveViewer

        editor = self.dialog.editor_frame
        range_id = self.range_id.get()
        ft = self.range_id_dict[range_id]
        scaled_y = self.dialog.mapper.make_scaled_xray_curve_y()
        viewer = ScatteringCurveViewer(editor, editor, self.sd, ft, scaled_y)
        viewer.show()



"""
    Selective.LrfSource.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import os
import logging
import numpy as np
from importlib import reload
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting, setting_file, save_settings
from molass_legacy.Models.ElutionCurveModels import EGH, EGHA
from molass_legacy.Models.ModelUtils import compute_area_props

NUM_MONOPOREPARAMS = 6

class LrfSource:
    def __init__(self, sd, corrected_sd, lrf_src_args1, peak_params_set, pre_recog=None):
        self.logger = logging.getLogger(__name__)
        self.sd = sd
        self.corrected_sd = corrected_sd
        uv_x, uv_y, xr_x, xr_y, baselines = lrf_src_args1
        self.uv_x = uv_x
        self.uv_y = uv_y
        self.xr_x = xr_x
        self.xr_y = xr_y
        self.baselines = baselines
        self.peak_params_set = peak_params_set
        self.uv_peaks = peak_params_set.uv_peaks
        self.xr_peaks = peak_params_set.xr_peaks
        if self.xr_peaks.shape[1] == 4:
            self.model = EGH()
        else:
            self.model = EGHA() 
        self.rg_info = None             # _compute_rgs call will be delayed
        self.egh_moments_list = None
        self.pre_recog = pre_recog
   
    def _compute_rgs(self,
                     keep_num_components=False,
                     want_num_components=False,
                     select=None, debug=False):
        if debug:
            import molass_legacy.Selective.LrfRgComputer
            reload(molass_legacy.Selective.LrfRgComputer)
        from .LrfRgComputer import compute_rgs_from_lrf_source
        if debug:
            print("Computing Rgs...")
        rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = compute_rgs_from_lrf_source(self,
                                                                                                          keep_num_components=keep_num_components,
                                                                                                          want_num_components=want_num_components,
                                                                                                          select=select)
        self.num_components_kept = keep_num_components
        if debug:
            def cb(ax):
                axt = ax.twinx()
                axt.grid(False)
                axt.plot(trs, rgs, 'o', color='green')
            self.draw(cb=cb, extra=False)
        self.rg_info = rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities

    def get_peaks(self):
        if self.rg_info is None:
            # note that this is intended to be used without any Rg info filtering
            return self.xr_peaks
        else:
            indeces = self.rg_info[6]
            return self.xr_peaks[indeces]

    def get_egh_moments_list(self):
        from molass_legacy.Models.Stochastic.MomentUtils import compute_egh_moments
        if self.egh_moments_list is None:
            peaks = self.get_peaks()
            self.egh_moments_list = compute_egh_moments(peaks)
        return self.egh_moments_list

    def compute_rgs(self, keep_num_components=False, want_num_components=None, select=None, debug=False):
        if self.rg_info is None:
            self._compute_rgs(keep_num_components=keep_num_components,
                              want_num_components=want_num_components,
                              select=select)
        assert keep_num_components == self.num_components_kept          # reconsider the timing of LrfSource construction if this assert fails
        return self.rg_info

    def draw(self, cb=None, extra=True, extra_button_specs=None, axes_info=None):
        self.moments_demo_params = None

        def plot_curves(ax, x, y, peaks, baseline, color=None):
            model = self.model
            ax.plot(x, y, color=color, label='data')
            cy_list = []
            for k, params in enumerate(peaks):
                cy = model(x, params)
                cy_list.append(cy)
            props = compute_area_props(cy_list)
            for k, (cy, prop) in enumerate(zip(cy_list, props)):
                ax.plot(x, cy, ":", label='component-%d (%.3g)' % (k, prop))
            ty = np.sum(cy_list, axis=0)
            ax.plot(x, ty, ":", color="red", label='model total')
            ax.plot(x, baseline, color="red", label='baseline')
            ax.legend()

        if axes_info is None:        
            if extra_button_specs is None:
                return_ret_simply = True
                extra_button_specs = [
                                    ("Guess Monopore Params", self.guess_monopore_params)
                                    ("Guess Lnpore Params", self.guess_lnpore_params),
                                    ("Moments Demo", self.moments_demo),
                                    ("Monopore Study", self.monopore_study),
                                    ("Lnpore Study", self.lnpore_study),
                                    ] if extra else []
            else:
                return_ret_simply = False

            with plt.Dp(button_spec=["OK", "Cancel"],
                        extra_button_specs=extra_button_specs):
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("LRF source")
                ax1.set_title("UV")
                plot_curves(ax1, self.uv_x, self.uv_y, self.uv_peaks, self.baselines[0], color='blue')
                ax2.set_title("XR")
                plot_curves(ax2, self.xr_x, self.xr_y, self.xr_peaks, self.baselines[1], color='orange')

                if cb is not None:
                    cb(ax2)

                fig.tight_layout()
                ret = plt.show()
        else:
            ax1, ax2 = axes_info
            plot_curves(ax1, self.uv_x, self.uv_y, self.uv_peaks, self.baselines[0], color='blue')
            plot_curves(ax2, self.xr_x, self.xr_y, self.xr_peaks, self.baselines[1], color='orange')
            return

        if return_ret_simply:
            return ret
        else:
            if ret:
                return self.moments_demo_params
            else:
                return
 
    def adapt_params(self, rgs, trs, props, indeces, peak_rgs, params, debug=False):
        # task: clarify the reason for len(rgs) > len(props)
        nc = len(props)
        print("adapt_params nc=", nc, "indeces=", indeces)
        N, T, x0, me, mp, poresize = params[0:NUM_MONOPOREPARAMS]
        scales = params[NUM_MONOPOREPARAMS:]
        ret_scales = np.zeros(nc)
        ret_scales[indeces] = scales
        whole_scale = np.sum(scales)
        for i in sorted(set(np.arange(nc)).difference(set(indeces))):
            ret_scales[i] = props[i]/whole_scale
        ret_params = np.array([x0, poresize, N, me, T, mp]), rgs[0:nc], ret_scales
        return ret_params

    def guess_monopore_params(self, old_style_return=False, debug=False):
        import molass_legacy.Models.Stochastic.MonoporeMoments as module
        reload(module)
        from molass_legacy.Models.Stochastic.MonoporeMoments import study_monopore_moments_impl
        ret = study_monopore_moments_impl(self,
                                          keep_num_components=True,     # this is required for backward compatibility
                                          # trust_max_num=2,              # do not trust Rg's exceeding this number to get possibly better fit
                                          debug=debug)
        if ret is None:
            return
        mnp_params, temp_rgs, unreliable_indeces, params_scaler = ret
        corrected_rgs = temp_rgs
        if len(unreliable_indeces) > 0:
            peak_rgs = self.rg_info[3]
            self.logger.info("peak_rgs=%s are replaced by corrected_rgs=%s with unreliable_indeces=%s", peak_rgs, corrected_rgs, unreliable_indeces)

        if old_style_return:
            N, T, t0, me, mp, poresize = mnp_params[0:6]
            # change the order of parameters to keep backward compatibility
            seccol_params = np.array([t0, poresize, N, me, T, mp])
            print("guess_monopore_params: ", seccol_params, corrected_rgs, mnp_params[6:])
            return seccol_params, corrected_rgs, mnp_params[6:]
        else:
            return mnp_params, corrected_rgs

    def guess_lnpore_params(self, return_rgs=False, use_study=True, use_moments=True, progress_cb=None, debug=False):
        if use_study:
            import molass_legacy.Models.Stochastic.MomentsStudy
            reload(molass_legacy.Models.Stochastic.MomentsStudy)
            from molass_legacy.Models.Stochastic.MomentsStudy import moments_study_impl
            return moments_study_impl(self, return_rgs=True, debug=debug)
        else:
            if debug:
                import molass_legacy.Models.Stochastic.MomentUtils
                reload(molass_legacy.Models.Stochastic.MomentUtils)
            from molass_legacy.Models.Stochastic.MomentUtils import compute_egh_moments
            from molass_legacy.Models.Stochastic.MonoporeGuess import guess_monopore_params_using_moments
            from molass_legacy.Models.Stochastic.LognormalGuess import guess_lognormalpore_params_using_moments
            rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = self.compute_rgs(debug=False)
            x, y = self.xr_x, self.xr_y
            model = self.model
            peaks = self.xr_peaks[indeces,:]
            egh_moments_list = compute_egh_moments(peaks)
            if use_moments:
                from molass_legacy.Models.Stochastic.RoughGuess import guess_monopore_params_roughtly
                monopore_params = guess_monopore_params_roughtly(x, y, model, peaks, peak_rgs, props, egh_moments_list)
                self.logger.info("monopore_params in guess_lnpore_params: %s", monopore_params)
                if progress_cb is not None:
                    progress_cb(1)
            else:
                from molass_legacy.Models.Stochastic.SecModelUtils import guess_monopore_params_from_rgdist
                monopore_params = guess_monopore_params_from_rgdist(x, y, peak_rgs, peak_trs, props, debug=False)   
        
            better_monopore_params = guess_monopore_params_using_moments(x, y, egh_moments_list, peak_rgs, props, monopore_params, debug=debug)
            self.logger.info("better_monopore_params in guess_lnpore_params: %s", better_monopore_params)
            if progress_cb is None:
                progress_cb_info = None
            else:
                progress_cb(2)
                progress_cb_info = (2, 10, progress_cb)
            lognormalpore_params = guess_lognormalpore_params_using_moments(x, y, egh_moments_list, peak_rgs, props, better_monopore_params,
                                                                            progress_cb_info=progress_cb_info,
                                                                            debug=debug)
            if progress_cb is not None:
                progress_cb(10)
            if return_rgs:
                return lognormalpore_params, peak_rgs
            else:
                return lognormalpore_params
    
    def moments_demo(self, adapt_backward=True, debug=True):
        if debug:
             import molass_legacy.Models.Stochastic.MomentUtils
             reload(molass_legacy.Models.Stochastic.MomentUtils)
        from molass_legacy.Models.Stochastic.MomentUtils import moments_demo_from_rgdist
        rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = self.compute_rgs(debug=False)
        # monopore_params = guess_monopore_params_from_rgdist(self.xr_x, self.xr_y, peak_rgs, peak_trs, props, debug=True)
        monopore_params = None
        demo_ret = moments_demo_from_rgdist(self.xr_x, self.xr_y, self.model, self.xr_peaks[indeces,:], peak_rgs, props, monopore_params,
                                            logger=self.logger, debug=True)
        if demo_ret is not None:
            ret_params = demo_ret[0]
            if adapt_backward:
                ret_params = self.adapt_params(rgs, trs, orig_props, indeces, peak_rgs, ret_params, debug=debug)
            self.moments_demo_params = ret_params

    def monopore_study(self, debug=True):
        if debug:
             import molass_legacy.Models.Stochastic.MonoporeStudy
             reload(molass_legacy.Models.Stochastic.MonoporeStudy)
        from molass_legacy.Models.Stochastic.MonoporeStudy import study
        rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = self.compute_rgs(debug=False)
        study(self.xr_x, self.xr_y, self.baselines[1], self.model, self.xr_peaks[indeces,:], peak_rgs, props)

    def lnpore_study(self, debug=True):
        if debug:
             import molass_legacy.Models.Stochastic.LnporeStudy
             reload(molass_legacy.Models.Stochastic.LnporeStudy)
        from molass_legacy.Models.Stochastic.LnporeStudy import study
        rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = self.compute_rgs(debug=False)
        study(self.xr_x, self.xr_y, self.baselines[1], self.model, self.xr_peaks[indeces,:], peak_rgs, props)

    def run_simulation_process(self, parent, debug=True):
        import multiprocessing as mp
        from molass_legacy._MOLASS.Processes import register_process
        if debug:
             import molass_legacy.Simulative.SimulationBridge
             reload(molass_legacy.Simulative.SimulationBridge)
        from molass_legacy.Simulative.SimulationBridge import simulation_bridge
        print("run_simulation_process")

        # setting info to pass to the simulation process
        analysis_folder = get_setting('analysis_folder')
        self.run_setting_file = os.path.join(analysis_folder, setting_file) 
        save_settings(self.run_setting_file)

        self.simulation_process = mp.Process(target=simulation_bridge, args=(self, ))
        register_process(self.simulation_process)
        self.simulation_process.start()


# coding: utf-8
"""
    ThreeDimUtils.py

    Copyright (c) 2018-2021, SAXS Tam, KEK-PF
"""
import numpy                as np

def compute_plane( A, B, C, x, y ):
    """
    base_plane = [
        [ A * x[0] + B * y[0], A * x[0] + B * y[1], A * x[0] + B * y[2] ],
        [ A * x[1] + B * y[0], A * x[1] + B * y[1], A * x[1] + B * y[2] ],
        [ A * x[2] + B * y[0], A * x[2] + B * y[1], A * x[2] + B * y[2] ],
        ...
        ] + C
                = np.dot( [
                            [ A * x[0], 1 ],
                            [ A * x[1], 1 ],
                            [ A * x[2], 1 ],
                              ...
                          ],
                          [
                            [ 1,        1,        1,        ... ],
                            [ B * y[0], B * y[1], B * y[2], ... ].
                            ...
                          ] )

    """
    return np.dot( np.vstack( [ A * x, np.ones( len(x) ) ] ).T, np.vstack( [ np.ones( len(y) ), B * y ] ) ) + C

def neaten_axes(axes):
    axis_lims = []
    for ax in axes:
        axis_lims.append((ax.get_xlim(), ax.get_ylim(), ax.get_zlim()))
    axis_lim_array = np.array(axis_lims)
    axis_limits = []
    for j  in range(3):
        vmin = np.min(axis_lim_array[:,j,0])
        vmax = np.max(axis_lim_array[:,j,1])
        axis_limits.append((vmin, vmax))
    for ax in axes:
        ax.set_xlim(*axis_limits[0])
        ax.set_ylim(*axis_limits[1])
        ax.set_zlim(*axis_limits[2])



"""
    Peaks.PeakFronting.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

def has_fronting_peak(ecurve, debug=False):
    x = ecurve.x
    y = ecurve.y
    j = ecurve.max_j
    hy = ecurve.max_y/2
    hyw = np.where(y > hy)[0]
    gap = np.where(hyw[1:] - hyw[:-1] > 1)[0]
    if len(gap) > 0:
        # task: retry to find a range with no gap
        ret_judge = False
    else:
        lenL = j - hyw[0]
        lenR = hyw[-1] - j
        ret_judge = lenL > lenR

    if debug:
        print("gap=", gap)
        if len(gap) == 0:
            print("lenL, lenR=", lenL, lenR)
        print("ret_judge=", ret_judge)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("Fronting Peak Detection")
            ax.plot(x, y, label="data")
            ax.plot(x[hyw], y[hyw], 'o', label="upper half")
            fig.tight_layout()
            plt.show()

    return ret_judge


# coding: utf-8
"""
    Error.ErrorModel.py

    based on 2017, Steffen M. Sedlak, et al.
    Quantitative evaluation of statistical errors in smallangle X-ray scattering measurements

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from scipy.optimize import minimize, curve_fit
from scipy.interpolate import UnivariateSpline
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Peaks.ElutionModels import egh

def error_model(q, Iq, k, c, Iarb):
    return 1/(k*q)*(Iq + 2*c*Iarb/(1 - c))      # Eq.(10) in the paper

K_SCALE = 1e6

class ErrorModel:
    def __init__(self, qv, D, E, spline=None):
        self.E = E

        i = bisect_right(qv, 0.02)
        cy = np.average(D[i-5:i+6,:], axis=0)
        j = np.argmax(cy)
        sy = np.average(D[:,j-5:j+6], axis=1)
        se = np.average(E[:,j-5:j+6], axis=1)

        slice_ = slice(10, -10)
        # slice_ = slice(None, None)
        qv_ = qv[slice_]
        sy_ = sy[slice_]
        se_ = se[slice_]
        se2 = se_**2

        def obj_func(p):
            k, c = p
            y_ = error_model(qv_, sy_, k*K_SCALE, c, sy[i])
            return np.sum(((y_ - se2)/y_)**2)

        ret = minimize(obj_func, (1, 0.5), method="Nelder-Mead")
        k_, c_ = ret.x
        print("k_, c_=", k_, c_)
        print("ret.success=", ret.success)
        print("ret.nit=", ret.nit)

        sem =  error_model(qv_, sy_, k_*K_SCALE, c_, sy[i])

        if False:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_yscale("log")
                ax.plot(qv_, se_)
                ax.plot(qv_, np.sqrt(sem))
                fig.tight_layout()
                plt.show()

        if spline is None:
            spline = UnivariateSpline(qv, sy, s=0, ext=3)

        self.spline = spline
        self.Iarb = sy[i]
        self.params = ret.x

    def get_error_curve(self, qv, scale=1):
        k, c = self.params
        return np.sqrt(error_model(qv, scale*self.spline(qv), k*K_SCALE, c, scale*self.Iarb))

def demo_impl(in_folder, v2_copy):
    from molass_legacy.DENSS.DenssUtils import fit_data_bc
    from DataUtils import get_in_folder
    from MatrixData import simple_plot_3d
    from molass_legacy.KekLib.BasicUtils import Struct

    D, E, qv, ecurve = v2_copy.get_xr_data_separate_ly()

    i = bisect_right(qv, 0.02)

    cx = np.arange(D.shape[1])
    cy = np.average(D[i-5:i+6,:], axis=0)

    p_init = ecurve.get_emg_peaks()[ecurve.primary_peak_no].get_params()
    popt, pcov = curve_fit(egh, cx, cy, p_init)
    cym = egh(cx, *popt)

    j = np.argmax(cy)
    sy = np.average(D[:,j-5:j+6], axis=1)
    ey = np.average(E[:,j-5:j+6], axis=1)

    fq, fy, fe, _ = fit_data_bc(qv, sy, ey)
    spline = UnivariateSpline(fq, fy, s=0, ext=3)

    em = ErrorModel(qv, D, E, spline=spline)
    eym = em.get_error_curve(qv)

    with plt.Dp():
        in_folder = get_in_folder(in_folder)
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(15,5))
        fig.suptitle("Error Model for %s based on 2017, Steffen M. Sedlak, et al." % in_folder, fontsize=20)
        ax1.set_title("Elution at Q=%g" % 0.02, fontsize=16)
        ax2.set_title("Scattering Intensity at j=%d" % j, fontsize=16)
        ax3.set_title("Scattering Error at j=%d" % j, fontsize=16)

        for ax in [ax2, ax3]:
            ax.set_yscale("log")
        ax1.plot(cx, cy, label="data")
        ax1.plot(cx, cym, label="egh")
        ax1.legend()

        ax2.plot(qv, sy, label="data")
        ax2.plot(fq, fy, label="P. B. Moore 1980")
        ax2.legend()

        ax3.plot(qv, ey, label="data")
        ax3.plot(qv, eym, label="model")
        ax3.legend()

        fig.tight_layout()
        plt.show()

    def demo2_impl():
        print("popt=", popt)
        h = popt[0]

        error_curves = []
        for k in cx:
            scale = cym[k]/h
            error_curves.append(em.get_error_curve(qv, scale=scale))

        ME = np.array(error_curves).T
        print(D.shape, ME.shape)

        if False:
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(10,4))
                ax1.plot(cx, cym)
                ax2.plot(cx, ME[i,:])
                fig.tight_layout()
                plt.show()

        fym = spline(qv)
        P_list = []
        ratio = 1/fym[i]
        MD = ratio*fym[:,np.newaxis] @ cym[np.newaxis,:]
        print("MD.shape=", MD.shape)

        with plt.Dp():
            fig = plt.figure(figsize=(15,10))
            # fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(15,5), subplot_kw=dict(projection="3d"))
            gs = GridSpec(2, 3)
            ax1, ax2, ax3 = [fig.add_subplot(gs[0,j], projection="3d") for j in range(3)]
            ax4, ax5, ax6 = [fig.add_subplot(gs[1,j]) for j in range(3)]
            simple_plot_3d(ax1, D, x=qv)
            simple_plot_3d(ax2, MD, x=qv)
            simple_plot_3d(ax3, ME, x=qv)
            fig.tight_layout()
            plt.show()

        exit()

        Mp_list = []
        Ma_list = []
        P_list = []
        rank = 1
        for n in range(100):
            E_ = ME * np.random.normal(0, 1, D.shape)
            M = MD + E_

            Mp_list.append(M)

            if False:

                with plt.Dp():
                    fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(15,4))
                    ax3.set_yscale("log")
                    ax1.plot(cx, cym, label="start model")
                    ax1.plot(cx, M[i,:], label="perturbated")
                    ax1.legend()
                    for ax in [ax2, ax3]:
                        ax.plot(qv, fym, label="start model")
                        ax.plot(qv, M[:,j], label="perturbated")
                        ax.legend()
                    fig.tight_layout()
                    plt.show()

            U, s, VT = np.linalg.svd(M)
            M_ = U[:,0:rank] @ np.diag(s[0:rank]) @ VT[0:rank,:]
            Ma_list.append(M_)

            c = M_[i,:]
            C = np.array([c])
            Cinv = np.linalg.pinv(C)
            P = M_ @ Cinv
            P_list.append(P[:,0])

        Mp_std = np.std(Mp_list, axis=0)
        Ma_std = np.std(Ma_list, axis=0)

        with plt.Dp():
            fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(15,5), subplot_kw=dict(projection="3d"))
            simple_plot_3d(ax1, ME, x=qv)
            simple_plot_3d(ax2, Mp_std, x=qv)
            simple_plot_3d(ax3, Ma_std, x=qv)
            fig.tight_layout()
            plt.show()

        P_std = np.std(P_list, axis=0)
        P = P_list[0]
        """
        P = D @ X
        X = D+ @ P
        Pe = np.sqrt((E**2) @ (X**2))
        """
        Dinv = np.linalg.pinv(D)
        X = Dinv @ P[:,np.newaxis]
        Pe = np.sqrt((ME**2) @ (X**2))

        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_yscale("log")
            ax.plot(qv, ME[:,j]/sy[i], label="data")
            ax.plot(qv, Pe/P[i], label="fixed propagation")
            ax.plot(qv, P_std/P[i], label="Monte Carlo")
            ax.legend()
            fig.tight_layout()
            plt.show()

    return demo2_impl

def demo(in_folder, trimming=True, correction=True):
    from molass_legacy.Tools.SdUtils import get_sd

    sd = get_sd(in_folder, trimming=trimming, correction=correction)
    demo2_impl = demo_impl(in_folder, sd)
    demo2_impl()



# coding: utf-8
"""
    MatrixData.py

    Copyright (c) 2019-2025, SAXS Team, KEK-PF
"""

import numpy as np
import logging
from mpl_toolkits.mplot3d import Axes3D, proj3d
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from molass_legacy.KekLib.OurMatplotlib3D import Inset2Din3D
from molass.PlotUtils.MatrixPlot import compute_3d_xyz, simple_plot_3d, contour_plot

class MatrixData:
    def __init__(self, data):
        self.logger = logging.getLogger(__name__)
        self.data = data
        self.i = np.arange(data.shape[0])
        self.j = np.arange(data.shape[1])
        self.e_index = None
        self.e_curve = None
        self.i_slice = None
        self.j_slice = None
        self.inset = None

    def plot(self, ax=None, color=None, alpha=1, title=None, ec_color='orange', inset=True):
        import molass_legacy.KekLib.DebugPlot as plt
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            need_show = True
        else:
            need_show = False

        if title is not None:
            ax.set_title(title, fontsize=16, y=1.09)

        x = self.vector
        simple_plot_3d(ax, self.data, x=x, color=color, alpha=alpha)

        if self.e_curve is not None:
            pos = x[self.e_index]
            y_ = self.j
            x_ = np.ones(len(y_))*pos
            z_ = self.e_curve.sy
            ax.plot(x_, y_, z_, color=ec_color)
            for rec in self.e_curve.peak_info:
                top_j = rec[1]
                x_ = self.vector
                y_ = np.ones(len(x_))*top_j
                z_ = self.data[:,top_j]
                ax.plot(x_, y_, z_, color='green')

        fc_exists = self.j_slice is not None and self.j_slice.start is not None and self.j_slice.start > 0
        if inset and self.e_curve is not None and fc_exists:
            islice, jslice = self.get_inset_slice()
            try:
                self.inset = inset = Inset2Din3D(ax, [0.6, 0.7, 0.4, 0.3])
                axins = inset.get_axis()

                x = self.j[jslice]
                index = (islice.start + islice.stop)//2
                y = self.data[index, jslice]
                axins.plot(x, y)
                ylim = axins.get_ylim()
                axins.set_ylim(ylim)
                fx = self.j_slice.start
                axins.plot([fx, fx], ylim, ':', color='yellow', linewidth=3)

                x_ = self.vector[index]
                ymin = x[0]
                ymax = x[-1]
                zmin = np.min(y)
                zmax = np.max(y)
                xi_pos = [x_]*5
                yi_pos = [ymin, ymax, ymax, ymin, ymin]
                zi_pos = [zmin, zmin, zmax, zmax, zmin]
                ax.plot(xi_pos, yi_pos, zi_pos, color='cyan')

                p1 = (x_, ymin, zmax)
                p2 = (x_, ymax, zmin)
                inset.set_annotation_lines(p1, p2)
                inset.draw()
                inset.set_event_handler()   # holding inset as self.inset seems to avoid losing (disabling) handler
            except:
                from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
                etb = ExceptionTracebacker()
                print(etb)

        if need_show:
            fig.tight_layout()
            plt.show()

    def set_elution_curve(self, index, slice_, correct=False):
        self.e_index = index
        self.e_slice = slice_
        y = np.average(self.data[slice_,:], axis=0)
        if correct:
            from LPM import get_corrected
            y = get_corrected(y)
        self.e_y = y
        try:
            self.e_curve = ElutionCurve(y)
        except:
            # e.g., No peak!
            self.e_curve = None
            if True:
                from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
                etb = ExceptionTracebacker()
                print(etb)
        print('self.e_curve=', self.e_curve)

    def set_restriction(self, i_slice, j_slice):
        self.i_slice = i_slice
        self.j_slice = j_slice

    def get_sliced_data(self):
        # print(self.i_slice, self.j_slice)
        return self.data[self.i_slice, self.j_slice], self.vector[self.i_slice]

    def get_inset_slice(self):
        isize = self.data.shape[0]
        jsize = self.data.shape[1]
        iw = int(isize*0.1)
        jw = int(jsize*0.2)
        icenter = int(self.data.shape[0]*0.7)
        jcenter = self.j_slice.start
        istart = max(0, icenter-iw) 
        istop = min(isize, icenter+iw)
        jstart = max(0, jcenter-jw) 
        jstop = min(jsize, jcenter+jw)
        return slice(istart, istop), slice(jstart, jstop)



"""
    Models/Stochastic/LognormalStudy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.stats import lognorm, norm
import matplotlib.pyplot as plt



if __name__ == "__main__":
    # import seaborn as sns
    # sns.set()
    N = 1000
    """
    mode = exp(mu - sigma**2)
    log(mode) = mu - sigma**2
    """
    k = 0
    for mu, sigma, xmax in [(2, 0.5, 25), (4.5, 0.1, 300), (4.5, 0.2, 300)]:
        fig, ax = plt.subplots()
        ax.set_title("Comparison among Normal, Exp(Normal), Lognormal with %d samples" % N)
        ax.set_xlabel("Pore Size")
        ax.set_ylabel("Density")
        X = np.random.normal(mu, sigma, N)
        if k == 0:
            ax_ = ax
        else:
            ax_ = ax.twinx()
        ax.hist(X, bins=50, density=True, alpha=0.5, label="normal", color="C0")
        ax_.hist(np.exp(X), bins=100, density=True, alpha=0.5, label="exp(normal)", color="C1")
        ymin, ymax = ax.get_ylim()
        if k == 0:
            tx = -1*0.9 + xmax*0.1
            xmin = -1
        else:
            tx = -1*0.95 + xmax*0.05
            xmin = -20
        ty = ymin*0.7 + ymax*0.3
        dx = (xmax - xmin)*0.1
        dy = (ymax - ymin)*0.2
        ax.annotate( "$Normal(\mu=%g,\sigma=%g)$" % (mu, sigma), fontsize=16, xy=(tx, ty), xytext=(tx+dx, ty+dy),
                    ha='center', va='center',
                    arrowprops=dict( headwidth=5, width=0.5, color='black', shrink=0.05) )
        ax.annotate( "$Lognormal(\mu=%g,\sigma=%g)$" % (mu, sigma), fontsize=16, xy=(tx+4*dx, ty-dy), xytext=(tx+5*dx, ty),
                    ha='center', va='center',
                    arrowprops=dict( headwidth=5, width=0.5, color='black', shrink=0.05) )
        Y = np.random.lognormal(mu, sigma, N)
        ax_.hist(Y, bins=100, density=True, alpha=0.5, label="lognormal", color="C2")
        ax.set_xlim(xmin, xmax)
        x = np.linspace(xmin, xmax, 1000)
        ax.plot(x, norm.pdf(x, mu, sigma), label="normal pdf", color="C3")

        ax_.plot(x, lognorm.pdf(x, sigma, scale=np.exp(mu)), label="lognormal pdf", color="C4")
        mode = np.exp(mu - sigma**2)
        ax.axvline(mode, color='red', alpha=0.5, linestyle='--', label="$mode=exp(\mu-\sigma^2)$")
        ax.legend()
        if k > 0:
            ax_.legend(loc='center right')
        fig.tight_layout()
        plt.show()
        k += 1





"""
    Kratky.GuinierKratkyInfo.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from scipy.stats import linregress
from molass_legacy.SerialAnalyzer.AnalyzerUtil import compute_conc_factor_util

class GuinierKratkyInfo:
    def __init__(self, optimizer, params, lrf_info, need_baseline=False):
        self.conc_factor = compute_conc_factor_util()   # consider a better place to do this

        separate_params = optimizer.split_params_simple(params)

        Pxr, Cxr, Puv, Cuv, mapped_UvD = lrf_info.matrices

        rg_params = lrf_info.get_valid_rgs(separate_params[2])
        self.rg_params = rg_params
        rg_ = np.average(rg_params)

        qv = optimizer.qvector

        q = 1.3/rg_
        i = bisect_right(qv, q*2)
        self.gslice = gslice = slice(0, i)
        self.qv2 = qv**2
        self.qv2_ = qv2_ = self.qv2[gslice]

        nc = len(rg_params)
        rgs = []
        I0s = []
        glny_s = []
        qrgs = []
        qrgnys = []
        nc_ = nc
        if need_baseline:
            nc_ += 1

        for k, cy in enumerate(Pxr.T[0:nc_]):
            glny_ = np.log(cy[gslice])
            slope, intercept = linregress(qv2_, glny_)[0:2]
            Rg = np.sqrt(-3*slope)
            rgs.append(Rg)
            I0 = np.exp(intercept)
            I0s.append(I0)
            qrg = qv*Rg
            qrgny = qrg**2*cy/I0
            glny_s.append(glny_ - intercept)
            qrgs.append(qrg)
            qrgnys.append(qrgny)

        self.rgs = rgs
        self.I0s = I0s
        self.Cuv = Cuv
        self.glny_s = glny_s
        self.qrgs = qrgs
        self.qrgnys = qrgnys

    def compute_adjacent_deviation_ratios(self):
        rgs = np.array(self.rgs)
        guinier_devs = np.abs(rgs[:-1] - rgs[1:])
        guinier_devs = guinier_devs/np.sum(guinier_devs)

        qrgnys = self.qrgnys
        kratky_devs = []
        for k, (y1, y2) in enumerate(zip(qrgnys[:-1], qrgnys[1:])):
            kratky_devs.append(np.sqrt(np.mean((y1 - y2)**2)))
        kratky_devs = np.array(kratky_devs)/np.sum(kratky_devs)

        devs = np.sqrt(guinier_devs**2 + kratky_devs**2)

        return devs/np.sum(devs)

    def get_molecular_masses(self):
        masses = []
        for k, i0 in enumerate(self.I0s):
            c = np.max(self.Cuv[k,:]) * self.conc_factor   # conc_factor ok?
            masses.append(i0/c)
        return masses



"""
    ExceptionTracebacker.py

    Copyright (c) 2016-2024, Masatsuyo Takahashi, KEK-PF
"""
import sys
import traceback
import logging

class ExceptionTracebacker:
    def __init__( self, call_stack=False ):
        ( exc_type_type_, exc_value, exc_traceback ) = sys.exc_info()
        self.c_seq = traceback.format_stack()[:-2] if call_stack else []
        self.e_seq = traceback.format_exception( exc_type_type_, exc_value, exc_traceback)[1:]

    def __repr__( self ):
        return ''.join( self.c_seq + self.e_seq )

    def log( self ):
        logger  = logging.getLogger( __name__ )
        logger.error( str(self) )

    def last_line( self ):
        return self.e_seq[-1][:-1]      # remove trailing '\n'

    def last_lines( self, n=2 ):
        return ''.join(self.e_seq[-n:])

def log_exception(logger, message, n=2):
    if logger is None:
        logger = logging.getLogger( __name__ )
    etb = ExceptionTracebacker()
    logger.warning(message + etb.last_lines(n=n))
    return etb

def warnlog_exception(logger, message, n=2):
    if logger is None:
        logger = logging.getLogger( __name__ )
    etb = ExceptionTracebacker()
    logger.warning(message + etb.last_lines(n=n).replace('Error', 'Warning'))
    return etb



"""
    Trimming.CdLimit.py

    Copyright (c) 2020-2025, SAXS Team, KEK-PF
"""
import numpy as np
import logging
from bisect import bisect_right
from scipy.optimize import minimize
from molass_legacy.KekLib.GeometryUtils import rotated_argmin
from molass_legacy._MOLASS.SerialSettings import get_setting

NUM_Q_POINTS = 10
CD_LIMIT_NOFIT_RATIO    = 0.02
CD_SCORE_OK_LIMIT       = 1.0
CD_LIMIT_FIND_ROTATION  = -np.pi/6

class CdLimit:
    def __init__(self, data, error, e_curve, qvector):
        self.logger = logging.getLogger( __name__ )
        self.data = data
        self.e_curve = e_curve
        self.qidx = bisect_right(qvector, 0.02)
        self.qindex = bisect_right(qvector, get_setting('cd_eval_qmax'))
        print('qindex=', self.qindex)
        self.aslice = slice(0, self.qindex)
        self.x = qvector[self.aslice]

    def get_limit(self, smooth_bq=True, debug=False):
        limit, surely = self.get_limit_impl(NUM_Q_POINTS, smooth_bq, debug=debug)
        if limit == NUM_Q_POINTS - 1:
            self.logger.info("retry due to an end limit.")
            limit, surely = self.get_limit_impl(NUM_Q_POINTS*2, smooth_bq, retry_argmin=True, debug=debug)
        return limit, surely

    def get_limit_impl(self, num_points, smooth_bq, retry_argmin=False, debug=False):
        if debug:
            import molass_legacy.KekLib.DebugPlot as plt

        start = max(1, int(len(self.x)*CD_LIMIT_NOFIT_RATIO))

        e_curve = self.e_curve
        pno = e_curve.primary_peak_no
        rec = e_curve.peak_info[pno]
        f, p, t = rec
        eslice = slice(p,t+1)
        c = e_curve.y[eslice]
        c = c/np.max(c)
        C = np.array([c, c**2])
        Cpinv = np.linalg.pinv(C)
        ok_info_list = []
        init_scales = []
        scores = []
        scales = []
        ab_pairs = []
        rank = 2
        for i in range(num_points):
            aslice = slice(i,self.qindex)
            M = self.data[aslice,eslice]
            U, s, VT = np.linalg.svd(M)
            M_ = U[:,0:rank] @ np.diag(s[0:rank]) @ VT[0:rank,:]
            P = M_ @ Cpinv
            a = P[:,0]
            b = P[:,1]
            ab_pairs.append((a,b))

            def min_norm_diff_bq_scaled_aq(pv):
                return np.sum((b - pv[0]*a)**2)

            init_scale = s[1]/s[0]
            ret = minimize(min_norm_diff_bq_scaled_aq, (init_scale,))
            scale = ret.x[0]
            score = np.sqrt(np.average((b - scale*a)**2))*100/a[self.qidx]
            init_scales.append(init_scale)
            scales.append(scale)
            scores.append(score)
            if score < CD_SCORE_OK_LIMIT:
                ok_info_list.append((i, score))

        if len(ok_info_list) > 0:
            limit = rotated_argmin(CD_LIMIT_FIND_ROTATION, scores, debug)
            if retry_argmin:
                if limit == num_points - 1:
                    m = np.argmin(scores[0:-1])     # excluding the end point
                    self.logger.info("the end limit %d has been replaced by no rotation argmin %d.", limit, m)
                    limit = m
            surely = False
        else:
            limit = 0
            ii = np.argmax(init_scales)
            il = np.argmin(scales)
            surely = ii == 0 and il == 0

        if debug:
            from matplotlib.gridspec import GridSpec
            from molass_legacy.KekLib.SciPyCookbook import smooth
            from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

            plt.push()
            rsize = 4
            ncols = 5
            nrows = num_points//ncols
            fig = plt.figure(figsize=(23,12))
            gs = GridSpec(nrows*rsize, ncols)
            main_axes = []
            diff_axes = []
            for m in range(nrows):
                main_row = []
                diff_row = []
                m_ = m*rsize
                for n in range(ncols):
                    main_ax = fig.add_subplot(gs[m_:m_+3,n])
                    diff_ax = fig.add_subplot(gs[m_+3,n])
                    main_row.append(main_ax)
                    diff_row.append(diff_ax)
                main_axes.append(main_row)
                diff_axes.append(diff_row)

            main_axes = np.array(main_axes)
            diff_axes = np.array(diff_axes)

            fig.suptitle("A(q), B(q) Debug Plots for " + get_in_folder(), fontsize=20)

            def draw_score_text(ax, score, color, dy=0):
                xmin, xmax = ax.get_xlim()
                ymin, ymax = ax.get_ylim()
                tx = (xmin + xmax)/2
                ty = (ymin + ymax)/2 + (ymax-ymin)*dy
                score_text = 'SCD=%.3g' % score
                ax.text(tx, ty, score_text, color=color, alpha=0.3, fontsize=40, ha='center', va='center')

            smooth_scales = []
            smooth_scores = []

            smooth_bq=True

            for i in range(num_points):
                a, b = ab_pairs[i]
                scale = scales[i]
                score = scores[i]
                m, n = divmod(i,ncols)
                ax = main_axes[m,n]
                x = self.x[i:]
                ax.grid(False)
                ax.get_xaxis().set_visible(False)
                ax.plot(x, a, color='C1')
                axt = ax.twinx()
                axt.plot(x, b, color='pink', label='B(q)')
                if smooth_bq:
                    sb = smooth(b)
                    axt.plot(x, sb, color='cyan', label='smooth B(q)')
                axt.plot(x, scale*a, color='blue', label='scale*A(q)')
                axt.legend()
                dy = 0.1 if smooth_bq else 0
                draw_score_text(ax, score, 'pink', dy=0.1)
                if smooth_bq:
                    def min_norm_diff_sbq_scaled_aq(pv):
                        return np.sum((sb - pv[0]*a)**2)
                    init_scale = init_scales[i]
                    ret = minimize(min_norm_diff_sbq_scaled_aq, (init_scale,))
                    smooth_scale = ret.x[0]
                    smooth_scales.append(smooth_scale)
                    sy = sb - smooth_scale*a
                    smooth_score = np.sqrt(np.average((sb - smooth_scale*a)**2))*100/a[self.qidx]
                    smooth_scores.append(smooth_score)
                    draw_score_text(ax, smooth_score, 'cyan', dy=-0.1)

                axd = diff_axes[m,n]
                y = b - scale*a
                axd.plot(x, y, color='gray', label='B(q) - scale*A(q)')
                if smooth_bq:
                    axd.plot(x, sy, color='cyan', label='smooth B(q) - scale*A(q)')

                axd.legend(bbox_to_anchor=(1, 1), loc='lower right')

            fig.tight_layout()
            fig.subplots_adjust(top=0.92)
            plt.show()
            plt.pop()

            a, b = ab_pairs[limit]
            scale = scales[limit]
            print('limit=', limit)
            plt.push()
            gs = GridSpec(4, 3)
            fig = plt.figure(figsize=(23,8))
            ax1 = fig.add_subplot(gs[0:2,0])
            ax2 = fig.add_subplot(gs[2:4,0])
            ax3 = fig.add_subplot(gs[0:3,1])
            ax3.get_xaxis().set_visible(False)
            ax3.grid(False)
            ax3d = fig.add_subplot(gs[3,1])
            ax4 = fig.add_subplot(gs[0:3,2])
            ax4.grid(False)
            ax4.get_xaxis().set_visible(False)
            ax4d = fig.add_subplot(gs[3,2])

            fig.suptitle("B(q) Limit Debug Plot for " + get_in_folder(), fontsize=20)
            ax1.set_title("Scales", fontsize=16)
            ax1.plot(init_scales, label='init scales')
            ax1.plot(scales, label='min norm scales')
            ax2.set_title("Scores", fontsize=16)
            if len(scores) > NUM_Q_POINTS:
                xticks = np.arange(0,len(scores), 2)
                for ax in [ax1, ax2]:
                    ax.set_xticks(xticks)
            ax2.plot(scores, label='scores')
            if smooth_bq:
                ax2.plot(smooth_scores, color='cyan', label='smooth scores')

            for ax in [ax1, ax2]:
                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymin, ymax)
                ax.plot([limit, limit], [ymin, ymax], ':', color='red', label='angular range start')
                ax.legend()

            x = self.x[limit:]
            ax3.set_title("B(q) - scale*A(q)", fontsize=16)
            ax3.plot(x, a, color='C1', label='A(q)')
            ax3.legend(loc='center right')
            ax3t = ax3.twinx()
            ax3t.plot(x, b, color='pink', label='B(q)')
            ax3t.plot(x, scale*a, color='blue', label='scale*A(q)')
            ax3t.legend()
            ax3d.plot(x, b - scale*a, color='gray', label='B(q) - scale*A(q)')
            ax3d.legend(bbox_to_anchor=(1, 1), loc='lower right')

            sb = smooth(b)
            ax4.set_title("smooth B(q) - scale*A(q)", fontsize=16)
            ax4.plot(x, a, color='C1', label='A(q)')
            ax4.legend(loc='center right')
            ax4t = ax4.twinx()
            ax4t.plot(x, sb, color='cyan', label='smooth B(q)')

            if smooth_bq:
                smooth_scale = smooth_scales[limit]
            else:
                def min_norm_diff_sbq_scaled_aq(pv):
                    return np.sum((sb - pv[0]*a)**2)
                init_scale = init_scales[i]
                ret = minimize(min_norm_diff_sbq_scaled_aq, (init_scale,))
                smooth_scale = ret.x[0]

            sa = smooth_scale*a
            ax4t.plot(x, sa, color='blue', label='scale A(q)')
            ax4t.legend()
            ax4d.plot(x, sb - sa, color='gray', label='smooth B(q) - scale*A(q)')
            ax4d.legend(bbox_to_anchor=(1, 1), loc='lower right')

            fig.tight_layout()
            fig.subplots_adjust(top=0.88)
            plt.show()
            plt.pop()

        return limit, surely



"""

    ファイル名：   PilatusImageInfoTable.py

    処理内容：

       Pilatus 画像データ情報の一覧

    Copyright (c) 2020,2024 SAXS Team, KEK-PF
"""
import sys
import copy
import logging
import traceback
from datetime               import datetime

from molass_legacy.KekLib.OurTkinter             import Tk, ToolTip
import OurMessageBox        as     MessageBox
from TkTableWrapper         import ArrayVar, Table
from ChangeableLogger       import Logger
from PilatusUtils           import get_data_info
from GuiPreferences         import PreferencesDialog
from PilatusImageProperty   import ImagePropertyWindow
from Preferences            import get_preference, temporary_preferences_begin, temporary_preferences_end, get_usual_preference
from SynthesizerSettings    import get_setting

# mapping displayed table col to internal array col
tif_dir_map = {
        1   :   'in_folder',
        2   :   'in_folder',
        5   :   'in_folder',
        8   :   'syn_folder',
        }

class PilatusImageInfoTable:
    # TODO: widget にすべきか否か
    def __init__( self, parent, synthesizer, suggestion, geometry_getter=None ):
        self.parent     = parent
        self.var        = None
        self.table      = None
        self.init_geometry = None
        self.current_row = 0
        self.synthesizer = synthesizer
        self.x_view     = None
        self.y_view     = None
        self.suggestion = suggestion
        self.geometry_getter = geometry_getter
        self.temporary_i_ratio_array = None
        self.yscrollbar_frame = None
        self.xscrollbar_frame = None
        self.current_data_end = 0

        self.refresh()

    def get_geometry(self):
        parent = self.parent
        parent.update()
        return parent.winfo_width(), parent.winfo_height()

    def refresh( self, restore_view=False, pilatus_counter=None, bottom_view=False, log_file_path=None, logger=None, autorun=False ):

        print('refresh: restore_view=', restore_view)
        print('refresh: bottom_view=', bottom_view)
        print('refresh: logger=', logger)

        temp_logger = Logger()

        if not restore_view:
            self.temporary_i_ratio_array = None

        self.num_selected_rows  = 0

        in_folder   = get_setting( 'in_folder' )
        adj_folder  = get_setting( 'adj_folder' )
        syn_folder  = get_setting( 'syn_folder' )

        counter_id = get_preference( 'detection_counter' )
        # print( 'counter_id=', counter_id )

        try:
            _, _, data_array, pilatus_counter = get_data_info( in_folder, adj_folder, syn_folder, pilatus_counter, counter_id,
                                                        log_file_path=log_file_path,
                                                        logger=temp_logger,
                                                        )
        except:
            logging.exception( 'Unexpected error' )
            MessageBox.showerror( 'Unexpected Error', traceback.format_exc() )
            return

        buffer = temp_logger.get_stream_buffer()
        if buffer.find("ERROR") > 0:
            if autorun:
                from ErrorLogCheck import all_known_errors
                show_message = not all_known_errors(buffer)
            else:
                show_message = True
            if show_message:
                    MessageBox.showerror( 'Refresh Error',
                                            "Following errors have been detected during refresh.\n"
                                            "Please check them before continuing.\n"
                                            "(You can copy this info with <Ctrl+c> to the clipboard\n"
                                            "in case needed)\n\n"
                                            + buffer.replace(",root", "") )

        new_num_rows = len(data_array)
        print( 'refresh: new_num_rows=', new_num_rows)

        self.data_array = data_array
        self.pilatus_counter = pilatus_counter

        if self.current_data_end > len( data_array ):
            # 作動中に出力フォルダのデータを削除した場合などが該当する。
            self.current_data_end = 0

        self.var = ArrayVar( self.parent )

        self.var['0,0'] = 'No'
        # self.var['0,1'] = 'Sample ID'       # TODO: change names derived from Image ID => Sample ID
        self.var['0,1'] = 'Original Image'
        # self.var['0,2'] = 'Absolute Position'
        self.var['0,2'] = 'Shifted Image 1'
        self.var['0,3'] = 'Relative Position 1'
        self.var['0,4'] = 'Intensity Ratio 1 %s' % ( counter_id )
        # self.var['0,5'] = 'Adjusted Image'
        self.var['0,5'] = 'Shifted Image 2'
        self.var['0,6'] = 'Relative Position 2'
        self.var['0,7'] = 'Intensity Ratio 2 %s' % ( counter_id )
        self.var['0,8'] = 'Synthesized Image'

        num_cols = 8

        row = 0
        for row_rec in self.data_array:
            row += 1
            # print( '[%d]' % (row), row_rec )
            self.var["%d,%d" % (row, 0)] = row
            sample_id   = row_rec[0]
            i = 0
            col = 0
            for sub_rec in row_rec[1]:
                i += 1

                col += 1
                if sub_rec[0] == None:
                    disp_value = ''
                else:
                    disp_value = sub_rec[0]
                self.var["%d,%d" % (row, col)] = disp_value

                if i == 1:
                    continue

                disp_value = '%s,%s' % tuple( sub_rec[1] )
                col += 1
                self.var["%d,%d" % (row, col)] = disp_value

                i_ratio_value = sub_rec[2]
                if i_ratio_value:
                    changed_counter_id = ''
                    if self.temporary_i_ratio_array:
                        # TODO
                        pass
                    disp_value = '%.5f%s' % ( i_ratio_value, changed_counter_id )   # 表示用の桁数調整
                else:
                    disp_value = ''
                col += 1
                self.var["%d,%d" % (row, col)] = disp_value

                if i == len( row_rec[1] ):
                    col = num_cols
                    if sub_rec[4] == None:
                        disp_value = ''
                    else:
                        disp_value = sub_rec[4]
                    self.var["%d,%d" % (row, col)] = disp_value

        need_resize = False
        if self.table is None:
            self.init_num_rows = len(self.data_array)
        else:
            self.save_xy_views()
            self.table.destroy()
            print('table destroyed')
 
        if self.init_num_rows == 0:
            print('must show invisible rows')
            need_resize = True

        self.table = Table( self.parent,
                     rows       = row + 1,
                     cols       = num_cols + 1,
                     state      = 'disabled',
                     # state      ='normal',
                     width      = 0,
                     maxwidth   = 1600,
                     # height     = 10,
                     height     = 21,
                     titlerows  = 1,
                     titlecols  = 1,
                     roworigin  = 0,
                     colorigin  = 0,
                     # selectmode = 'browse',
                     selectmode = 'extended',
                     selecttype = 'row',
                     rowstretch = 'unset',
                     # colstretch = 'last',
                     colstretch = 'all',
                     browsecmd  = self.browsecmd,
                     flashmode  = 'on',
                     variable   = self.var,
                     usecommand = 0,
                     command    = self.test_cmd,
                     colwidth   = 20
                     )

        self.num_selectable_cols = num_cols - 1

        # thanks to sof
        # http://effbot.org/zone/tkinter-scrollbar-patterns.htm
        if self.yscrollbar_frame:
            self.yscrollbar_frame.destroy()
        self.yscrollbar_frame = Tk.Frame( self.parent )
        self.yscrollbar_frame.pack(side=Tk.RIGHT, fill=Tk.Y)
        # self.yscrollbar_frame.grid( row=0, column=1 )
        self.yscrollbar = Tk.Scrollbar( self.yscrollbar_frame, orient='vertical', command=self.table.yview_scroll )
        self.table.config( yscrollcommand=self.yscrollbar.set )
        self.yscrollbar_isvisible = False

        if self.xscrollbar_frame:
            self.xscrollbar_frame.destroy()
        self.xscrollbar_frame = Tk.Frame( self.parent )
        self.xscrollbar_frame.pack(side=Tk.BOTTOM, fill=Tk.X)
        # self.xscrollbar_frame.grid( row=1, column=0 )
        self.xscrollbar = Tk.Scrollbar( self.xscrollbar_frame, orient='horizontal', command=self.table.xview_scroll )
        self.table.config( xscrollcommand=self.xscrollbar.set )
        self.xscrollbar_isvisible = False

        self.on_resize()

        self.table.pack( expand=1, fill=Tk.BOTH )
        # self.table.grid( row=0, column=0 )

        self.table.tag_configure('sel', background = 'steel blue')
        self.table.tag_configure('active', background = 'blue')
        # self.table.tag_configure('title', anchor='w', bg='skyblue', relief='sunken')
        self.table.tag_configure('title', bg='gray25', relief='sunken')
        self.table.tk.call( self.table._w, 'width',
            0,  5,      # No
            # 1, 16,      # Image ID
            1, 36,      # Original Image
            # 2, 20,      # Absolute Position
            2, 36,      # Shifted Image 1
            3, 16,      # Relative Position 1
            4, 18,      # Intensity Ratio 1

            5, 36,      # Shifted Image 2
            6, 16,      # Relative Position 2
            7, 18,      # Intensity Ratio 2

            5, 36,      # Adjusted Image

            8, 36,      # Synthesized Image
            )


        if restore_view:
            self.restore_xy_views()

        if bottom_view:
            self.set_view_to_bottom()
            self.on_resize()

        self.select_added_rows()

        print( 'refresh: after select_added_rows y_view=', self.y_view )
        print( 'refresh: after select_added_rows yscrollbar_isvisible=', self.yscrollbar_isvisible )
        print( 'refresh: after select_added_rows height=',  self.table.cget( 'height' ))
        if need_resize:
            print('on_resize when need_resize')
            # self.table.see('topleft')
            self.on_resize()

        self.table.see('end')

        if self.init_geometry is None:
            # this does not get what is wanted.
            self.init_geometry = self.get_geometry()
            print('init_geometry=', self.init_geometry)

        # create a popup menu
        self.menu = Tk.Menu( self.table, tearoff=0 )
        self.menu.add_command( label='Show Original Images',    command=lambda: self.do_action( 1 ) )
        self.menu.add_command( label='Show Adjusted Images',    command=lambda: self.do_action( 2 ) )
        self.menu.add_command( label='Make Synthsized Images',  command=lambda: self.do_action( 3 ) )
        self.menu_c = Tk.Menu( self.table, tearoff=0 )
        # ToolTip( self.menu_c, 'With the cascading submenus, you can temporarily change the preferences and execute.')
        # does not work expectedly
        self.menu.add_cascade( label='Execute with Temporary Preference Changes', menu=self.menu_c )
        self.menu_c.add_command( label='Show Original Simply',   command=lambda: self.do_action( 1, True ) )
        self.menu_c.add_command( label='Show Adjusted Images',   command=lambda: self.do_action( 2, True ) )
        self.menu_c.add_command( label='Make Synthsized Images', command=lambda: self.do_action( 3, True ) )
        self.menu.add_command( label='Tiff File Description',    command=self.show_properties )

        self.table.bind( "<Button-3>", self.popup )
        self.table.bind( "<Button-1>", self.popup_state_reset )
        # self.table.bind( "<Triple-Button-1>", lambda evet: self.do_action( 1 ) )
        self.table.bind( "<Double-Button-1>", lambda evet: self.do_action( 1 ) )
        self.table.bind( "<Configure>", self.on_resize )

        ToolTip( self.table, '<Double-click> a row to show images. Select and <right-click> to show the action menu. Click top-left corner "No" to select all rows.')

        return need_resize

    def get_resize_height(self):
        num_rows = min(20, len(self.data_array))
        height = int(self.table.cget('height'))
        return num_rows * height

    def on_resize( self, event=None ):
        f, t = self.table.yview()
        if f > 0 or t < 1:
            if not self.yscrollbar_isvisible:
                self.yscrollbar.pack(side=Tk.RIGHT, fill=Tk.Y)
                self.yscrollbar_isvisible = True
        else:
            if self.yscrollbar_isvisible:
                self.yscrollbar.pack_forget()
                self.yscrollbar_isvisible = False

        f, t = self.table.xview()
        if f > 0 or t < 1:
            if not self.xscrollbar_isvisible:
                self.xscrollbar.pack(side=Tk.BOTTOM, fill=Tk.X)
                self.xscrollbar_isvisible = True
        else:
            if self.xscrollbar_isvisible:
                self.xscrollbar.pack_forget()
                self.xscrollbar_isvisible = False
                # self.xscrollbar_frame.configure( height=0 )

    def popup( self, event ):
        self.popup_position = ( event.x, event.y )
        # print( 'popup_position=', self.popup_position )

        # 改造した Tktable.dll ならば次の work-around は不要
        # self.table.configure( state='normal' )

        self.menu.post( event.x_root, event.y_root )

    def popup_state_reset( self, event ):
        # 改造した Tktable.dll ならば継ぎの work-around は不要
        # self.table.configure( state='disabled' )
        pass

    def set_view_to_bottom( self ):
        height  = self.table.cget( 'height' )
        rows    = self.table.cget( 'rows' )
        # caveate: silent exception without int()!
        ypos = max( 0, 1 - int(height) / int(rows)  )
        if ypos < 0:
            ypos = 0
        self.table.yview_moveto( ypos )

    def test_cmd( self, event ):
        if event.i == 0:
            return '%i, %i' % (event.r, event.c)
        else:
            return 'set'

    def test_select_row( self, row ):
        # TODO: remove this method
        self.num_selected_rows = 1
        self.current_row = row + 1

    def set_num_selected_rows( self ):
        self.num_selected_rows = len( self.table.curselection() ) // ( self.num_selectable_cols + 1 )

    def browsecmd( self, event ):
        self.selection_array = []
        self.set_num_selected_rows()
        # print( 'self.num_selected_rows=', self.num_selected_rows )
        active_cell_index = self.table.index('active')
        # print( 'active_cell_index=', active_cell_index )
        self.current_row = int( active_cell_index.split(',')[0] )
        # print( 'self.current_row=', self.current_row )

    def get_last_cell_index( self ):
        num_rows = len( self.data_array )
        return '%d,%d' % ( num_rows,  self.num_selectable_cols + 1 )

    def select_all_rows( self ):
        last_cell = self.get_last_cell_index()
        self.table.selection_set( '1,1', last_cell )
        self.set_num_selected_rows()

    def select_added_rows( self ):
        if len( self.data_array ) == 0: return

        last_cell   = self.get_last_cell_index()

        if self.current_data_end < len( self.data_array ):
            first_cell  = '%d,1' % ( self.current_data_end + 1 )
            self.table.selection_set( first_cell, last_cell )
            self.set_num_selected_rows()
        else:
            self.table.selection_clear( '0,0', last_cell )

    def get_selected_indices( self ):
        curselection = self.table.curselection()
        # print( 'curselection=', curselection )
        indices = []
        for c in curselection[ slice( 0, len(curselection), self.num_selectable_cols + 1 ) ]:
            i = int( c.split(',')[0] ) - 1
            indices.append( i )
        # print( 'indices=', indices )
        return indices

    def get_not_yet_done_indices( self ):
        indices = []
        for k, row in enumerate(self.data_array):
            key, recs = row[0:2]
            syn_file = recs[0][4]
            if syn_file is None:
                indices.append(k)
        return indices

    def clear_selection( self, indices ):
        for i in indices:
            begin   = '%d,0' % ( i + 1 )
            end     = '%d,%d' % ( i + 1, self.num_selectable_cols + 1 )
            self.table.selection_clear( begin, end )

    def select_data_array( self, selected_indices ):
        array = []
        for i in selected_indices:
            array.append( self.data_array[ i ] )
        return array

    def do_action( self, action, change=False ):
        self.popup_state_reset( 0 )

        if self.suggestion and self.suggestion.is_blinking():
            MessageBox.showwarning( "Not Allowed", "You can't do any action until you refresh the image data infomation table." )
            return

        selected_indices = self.get_selected_indices()
        exec_array = self.select_data_array( selected_indices )

        if len( exec_array ) == 0:
            # this case do happen.
            MessageBox.showinfo( 'No Rows Selected Notification', 'No rows are selected. Select row(s) and retry.' )
            return

        counter_id_is_changed = False
        if change:
            temporary_preferences_begin()
            dialog = PreferencesDialog( self.parent, "Temporary Preferences", self.pilatus_counter,
                            action=action )
            if not dialog.applied:
                return

            counter_id = get_preference( 'detection_counter' )
            if counter_id != get_usual_preference( 'detection_counter' ):
                counter_id_is_changed = True
                counter_dict = self.pilatus_counter.get_counter_dict( counter_id )
                if not self.temporary_i_ratio_array:
                    self.temporary_i_ratio_array = []
                    for rec in self.data_array:
                        self.temporary_i_ratio_array.append( None )

        start = datetime.now()
        # print '%s start.' % ( start.strftime( '%Y-%m-%d %H:%M:%S' ) )

        self.synthesizer.execute( action, exec_array )

        end = datetime.now()
        t = end - start
        # print '%s end. took %d seconds' % ( end.strftime( '%Y-%m-%d %H:%M:%S' ), t.seconds )

        if change:
            temporary_preferences_end()

        if action == 3:
            self.update_current( selected_indices )
            self.refresh( restore_view=True )
            # TODO: update in synthesizer.single_exec

        return

    def update_current( self, selected_indices ):
        if len( selected_indices ) > 0:
            end_index = selected_indices[-1] + 1
            if end_index > self.current_data_end:
                self.current_data_end = end_index
                print( 'new self.current_data_end=', self.current_data_end )

    def dummy( self ):
        pass

    def show_properties( self ):
        self.popup_state_reset( 0 )
        curselection = self.table.curselection()
        if len( curselection ) == 0:
            # this case do happen.
            MessageBox.showinfo( 'No Row Selected Notification', 'No row is selected. Select a row and retry.' )
            return

        # TODO: also consider right-click position
        active_cell_index = self.table.index('active')
        print( "active cell index:", active_cell_index )

        first_row = int( curselection[0].split( ',' )[0] )
        col = int( active_cell_index.split(',')[1] )
        if col in tif_dir_map:
            dirkey = tif_dir_map[ col ]
            filename = self.var[ '%d,%d' % (first_row, col) ]
            dirpath = get_setting( dirkey )
            path = dirpath + '/' + filename
            title = filename + ' Description'
            ImagePropertyWindow( self.parent, title, path )
        else:
            MessageBox.showinfo( 'Select Tiff File Notice', 'Select tiff file cell and retry.' )
        return

    def save_xy_views( self ):
        self.x_view = self.table.xview()
        self.y_view = self.table.yview()

    def restore_xy_views( self ):
        if self.x_view and self.y_view:
            self.table.xview_moveto( self.x_view[0] )
            self.table.yview_moveto( self.y_view[0] )



"""
    V2PropOptimizer.RdrChart.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

def draw_rdr_chart(title, rdr, devel=True):
    indeces = [str(i) for i in range(len(rdr))]
    m = np.argmin(rdr)
    x = m
    y = np.max(rdr)*0.5
    dy = (np.min(rdr) - y)*0.8
    head_length = y*0.2
    with plt.Dp(window_title='RDR_AD Chart', ok_only=True, ok_text="Close"):
        fig, ax = plt.subplots()
        ax.set_title(title, fontsize=16)
        ax.bar(indeces, rdr)
        ax.arrow(x=x,y=y,dx=0,dy=dy, width=0.5, head_width=1, head_length=head_length,length_includes_head=True, color='red', alpha=0.5)
        fig.tight_layout()
        plt.show()



"""

    ElutionModelDemo.py

    Copyright (c) 2017-2023, SAXS Team, KEK-PF

"""
import numpy as np
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Mapping.SingleComponent import PEAK_EVAL_RANGE_RATIO

def demo(caller):
    print("ElutionModelDemo")
    editor = caller.editor
    opt_recs = editor.opt_recs
    print(len(opt_recs))

    rec = opt_recs[1]

    x = editor.x
    y = editor.y

    x_ = x - x[0]
    cy = rec.evaluator(x_)

    m = np.argmax(cy)
    xL = x[0:m]
    cyL = cy[0:m]
    xR = x[m:]
    cyR = cy[m:]

    with plt.Dp():
        fig = plt.figure(figsize=(12,6))
        fig.suptitle("EGHA, EMGA Inverse Spline Demo", fontsize=20)
        gs = GridSpec(7,2)
        ax0 = fig.add_subplot(gs[:,0])
        ax1 = fig.add_subplot(gs[0:4,1])
        ax2 = fig.add_subplot(gs[4:,1])

        ax0.set_title("Attending Element", fontsize=16)
        ax1.set_title("Left Side Inverse", fontsize=16)
        ax2.set_title("RIght Side Inverse", fontsize=16)

        ax0.plot(x, y)
        ax0.plot(xL, cyL, ":", lw=3)
        ax0.plot(xR, cyR, ":", lw=3)

        xmin, xmax = ax0.get_xlim()
        ax0.set_xlim(xmin, xmax)
        yf = cy[m]*PEAK_EVAL_RANGE_RATIO
        ax0.plot([xmin, xmax], [yf, yf], color="red")

        ax1.plot(cyL, xL, ":", lw=3, color="C1")
        ymin, ymax = ax1.get_ylim()
        ax1.set_ylim(ymin, ymax)
        ax1.plot([yf, yf], [ymin, ymax], color="red")

        ax2.plot(-cyR, xR, ":", lw=3, color="C2")
        ymin, ymax = ax2.get_ylim()
        ax2.set_ylim(ymin, ymax)
        ax2.plot([-yf, -yf], [ymin, ymax], color="red")

        fig.tight_layout()
        plt.show()



# -*- coding: utf-8 -*-
"""

    ファイル名：   SAnglerMask.py

    処理内容：

        SAngler マスクファイルのオブジェクト化。

"""
from __future__ import unicode_literals
import re
import numpy as np

comment_line_re = re.compile( '^#\s+(\w+)=(.+)' )

class SAnglerMask():
    def __init__( self, filename ):

        width, height   = 0, 0
        self.mask_array = []

        # print 'DEBUG: SAnglerMask(%s)' % ( filename )

        fh = open( filename )
        for line in fh.readlines():
            m1 = comment_line_re.match( line )
            if m1:
                key_word = m1.group( 1 )
                value    = m1.group( 2 )
                if key_word == 'Width':
                    width = int( value )
                elif key_word == 'Height':
                    height = int( value )
                continue

            if width == 0 or height == 0:
                raise Exception( 'Wrong format' )
            else:
                if len( self.mask_array ) == 0:
                    self.mask_array = np.zeros( [ height, width ] )

            for pair in line.split("\t"):
                (c, r) = pair.split()
                self.mask_array[ int(r), int(c) ] = 1

        fh.close()
        if width == 0 or height == 0:
            # ここは空ファイルなどが該当する。
            # マスクすべきピクセルが無い場合はありえても、Width と Height は必須とする。
            raise Exception( 'Wrong format' )



"""
    Optimizer.FuncImporter.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""
import os
import glob
import re
from importlib import import_module, reload

OBJFUNC_DIRNAME = "ObjectiveFunctions"

def import_objective_function(class_code, logger=None):
    try:
        import molass_legacy.Optimizer.BasicOptimizer
        reload(molass_legacy.Optimizer.BasicOptimizer)

        module = import_module("molass_legacy.%s.%s" % (OBJFUNC_DIRNAME, class_code))
        module = reload(module)
        class_ = getattr(module, class_code)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(logger, "importing ObjectiveFunctions: ", n=5)
        class_ = None
    return class_

def get_objective_function_info(logger=None, default_func_code=None, debug=False):
    # note that default_objective_func can be changed depending on elution model
    # so, making this to a singleton needs careful streatment of such changes

    if debug:
        import molass_legacy.Optimizer.BasicOptimizer as base_opt
        reload(base_opt)

    from molass_legacy._MOLASS.SerialSettings import get_setting
    from molass_legacy.KekLib.BasicUtils import Struct

    elution_model = get_setting('elution_model')
    if default_func_code is None:
        # this should have been set in OptStrategyDialog.py
        default_func_code = get_setting('default_objective_func')
    func_dict = {}
    key_list = []
    default_index = None
    file_re = re.compile(r'\W(\w\d+)\.py')
    upper_dir = os.path.dirname(os.path.dirname(__file__))
    if debug:
        print("upper_dir:", upper_dir)
    for k, file in enumerate(sorted(glob.glob(upper_dir + r"\%s\*.py" % OBJFUNC_DIRNAME))):
        if debug:
            print("importing objective function %d: %s" % (k, file))
        m = file_re.search(file)
        if m:
            class_code = m.group(1)
            if debug:
                print("class_code:", class_code)
            if elution_model == 0:
                if class_code >= "G0500":
                    continue
            elif elution_model == 1:
                if class_code < "G1000":
                    continue
            elif elution_model == 2:
                if class_code < "G0500" or class_code >= "G0600":
                    continue
            elif elution_model == 3:
                if class_code < "G0600" or class_code >= "G0700":
                    continue
            elif elution_model == 4:
                if class_code < "G0700" or class_code >= "G0800":
                    continue
            elif elution_model == 5:
                if class_code < "G2000" or class_code >= "G3000":
                    continue
            else:
                assert False

            if class_code == default_func_code:
                default_index = len(key_list)
            class_ = import_objective_function(class_code, logger)
            docstr = class_.__doc__
            key_str = ' : '.join([class_code, docstr])
            func_dict[key_str] = class_
            key_list.append(key_str)
            logger.info("function %s appended", class_code)

    func_info = Struct(func_dict=func_dict, key_list=key_list, default_index=default_index)

    return func_info



# coding: utf-8
"""
    SecTheory.SepseyProof.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.special import iv
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt

def fitting_demo_impl(in_folder, stochastic_model, init_params, niter):
    from scipy.optimize import curve_fit
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Baseline.BaselineUtils import get_corrected_sd_impl
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.Peaks.ElutionModels import egh
    from LPM import get_corrected
    from DataUtils import get_in_folder
    from SecSaxs.DataTreatment import DataTreatment
    from .StochasticSolver import CfDomain

    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)

    pre_recog = PreliminaryRecognition(sd)
    treat = DataTreatment(route="v2", trimming=2, correction=1)
    sd_copy = treat.get_treated_sd(sd, pre_recog)

    D, E, qv, ecurve = sd_copy.get_xr_data_separate_ly()

    fig, ax = plt.subplots()

    x = ecurve.x
    y = ecurve.y

    solver = CfDomain(x, y)

    for k in range(niter):
        opt_result = solver.fit_impl(stochastic_model, init_params=init_params)
        init_params = opt_result.x

    spline = solver.get_spline(init_params)

    ax.plot(x, y, label="data")
    ax.plot(x, spline(x), label="model")

    ax.legend()

    fig.tight_layout()
    plt.show()

def fitting_demo_lognormal_pore_cf(in_folder, init_params, niter=1):
    from .Sepsey2014 import lognormal_pore_cf
    fitting_demo_impl(in_folder, lognormal_pore_cf, init_params, niter)



# coding: utf-8
"""
    DataDebugger.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import os
import re
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry

foler_re = re.compile(r'(\d+)$')

class DataDebugger:
    def __init__(self, folder_name="debugger-00"):
        folder = os.path.join(get_setting('analysis_folder'), folder_name)
        folder = self.get_new_folder(folder)

        mkdirs_with_retry(folder)
        self.folder = folder

    def get_new_folder(self, folder):
        while os.path.exists(folder):
            folder = re.sub(foler_re, lambda m: '%02d' % (int(m.group(1))+1), folder)
            # print('next folder=', folder)

        # print('final folder=', folder)
        return folder

    def get_sub_folder(self, subfolder_name="00"):
        sub_folder = os.path.join(self.folder, subfolder_name)
        sub_folder = self.get_new_folder(sub_folder)
        mkdirs_with_retry(sub_folder)
        return sub_folder



"""
    Models/RateTheory/GRM.py

    Copyright (c) 2023, SAXS Team, KEK-PF





"""
import numpy as np

class GRM:
    def __init__(self, u, a, e, ep, D_L. Dp, Ds, Rp):
        self.u = u
        self.D_L = D_L
        self.F = (1 - e)/e
        self.Deff = ep*Dp + (1- ep)*Ds*a
        self.Rp = Rp
        self.a = a
        self.a_ = ep + (1 - ep)*a
        self.Bp = kext * Rp / self.Deff

    def alpha(self, s):
        return self.a_*s/self.Deff

    def f(self, s):
        ras = np.sqrt(self.alpha(s))
        return self.Bp /( self.Bp - 1 + ras*self.Rp / np.tanh(ras*self.Rp) )    # np.coth(...) == 1/np.tanh(...)

    def phi(self, s):
        return s + self.kext * 3/self.Rp * self.F * (1 - self.f(s))

    def c_(self, s, z):
        U = self.u / (2*self.D_L)
        V = np.sqrt(1 + 4*self.D_L*self.phi(s))
        lam1 = U*(1 - V)
        lam2 = U*(1 + V)
        return self.A * np.exp(lam1*z) + self.B * np.exp(lam2*z)

    def c_p(s, z, r):
        ras = np.sqrt(self.alpha(s))
        # return 1/r * ( self.d1(s,z)/ras * np.sinh(ras*r) + self.d2(s,z) * np.cosh(ras*r) )
        return 1/r * ( self.d1(s,z)/ras * np.sinh(ras*r)    # d2 == 0

    def d1(self, s, z, ras):
        ras = np.sqrt(self.alpha(s))
        nume = self.Bp * senl.c_(s, z) / np.sinh(ras*self.Rp)
        deno = (self.Bp - 1)/(ras*self.Rp) + 1 / np.tanh(ras*self.Rp
        return  nume/deno



"""
    Egh2Emg.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.Models.ElutionCurveModels import emga

def to_emga_params(params):
    ret_params = params.copy()
    tR = params[1]
    mu = tR - params[3]
    ret_params[1] = mu
    x = np.array([tR])
    y = emga(x, *ret_params)        # note that emga returns a vector
    ret_params[0] *= params[0]/y[0]
    return ret_params



# coding: utf-8
"""
    AverageSubtractorDialog.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import re
import logging
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk, is_empty_val
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable, BlinkingFrame
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
try:
    import molass_legacy.KekLib.CustomMessageBox         as MessageBox
except:
    import OurMessageBox            as MessageBox
from SerialDataUtils import get_xray_files
from DataSubtractor import subtract_impl
from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry
from SerialDataUtils import get_mtd_filename

class AverageSubtractorDialog(Dialog):
    def __init__(self, parent, use_mtd=True):
        self.logger = logging.getLogger(__name__)
        self.busy = False
        self.background_info = None
        self.in_folder_prev = None
        self.use_mtd = use_mtd
        Dialog.__init__(self, parent, "Average Subtractor", visible=False)

    def body(self, body_frame):
        tk_set_icon_portable(self)

        iframe = Tk.Frame(body_frame)
        iframe.pack(padx=20, pady=10)

        folder_entry_width = 60
        num_files_label_width = 6

        grid_row = -1

        grid_row += 1
        label = Tk.Label(iframe, text="Folder Path")
        label.grid(row=grid_row, column=1)
        label = Tk.Label(iframe, text="File Name Pattern")
        label.grid(row=grid_row, column=2, padx=10)
        label = Tk.Label(iframe, text="Files")
        label.grid(row=grid_row, column=3)

        grid_row += 1
        label = Tk.Label(iframe, text="Sample Data")
        label.grid(row=grid_row, column=0)

        self.in_folder1 = Tk.StringVar()
        self.in_files1 = []
        self.name_pattern1 = Tk.StringVar()
        self.num_files1 = Tk.IntVar()
        self.in_folder1_cb = lambda: self.on_entry_in_folder(self.in_folder1, self.in_files1, self.name_pattern1, self.num_files1)
        self.in_folder1_entry = FolderEntry(iframe, textvariable=self.in_folder1, width=folder_entry_width,
                                on_entry_cb=self.in_folder1_cb)
        self.in_folder1_entry.grid(row=grid_row, column=1)
        self.in_folder1_entry.bind('<Leave>', lambda *args:self.in_folder1_cb())

        label = Tk.Label(iframe, textvariable=self.name_pattern1)
        label.grid(row=grid_row, column=2, padx=10, sticky=Tk.W)

        num_files_label = Tk.Label(iframe, textvariable=self.num_files1, width=num_files_label_width, justify=Tk.RIGHT)
        num_files_label.grid(row=grid_row, column=3)

        grid_row += 1
        self.maker_btn_frame = BlinkingFrame(iframe)
        self.maker_btn_frame.grid(row=grid_row, column=0, pady=5)

        self.maker_btn = Tk.Button(self.maker_btn_frame, text="Backgroud Maker", command=self.show_backgroud_maker_dialog, state=Tk.DISABLED)
        self.maker_btn.pack()

        self.maker_btn_frame.objects = [self.maker_btn]

        self.in_folder2 = Tk.StringVar()
        self.in_files2 = []
        self.name_pattern2 = Tk.StringVar()
        self.num_files2 = Tk.IntVar()
        in_folder2_entry = FolderEntry(iframe, textvariable=self.in_folder2, width=folder_entry_width,
                                on_entry_cb=None)
        in_folder2_entry.grid(row=grid_row, column=1, pady=5)

        label = Tk.Label(iframe, textvariable=self.name_pattern2)
        label.grid(row=grid_row, column=2, padx=10, sticky=Tk.W)

        num_files_label = Tk.Label(iframe, textvariable=self.num_files2, width=num_files_label_width, justify=Tk.RIGHT)
        num_files_label.grid(row=grid_row, column=3)

        grid_row += 1
        label = Tk.Label(iframe, text="Subtracted Data")
        label.grid(row=grid_row, column=0)
        self.out_folder = Tk.StringVar()
        out_folder_entry = FolderEntry(iframe, textvariable=self.out_folder, width=folder_entry_width,
                                on_entry_cb=self.on_entry_out_folder)
        out_folder_entry.grid(row=grid_row, column=1)
        out_folder_entry.bind('<Leave>', lambda *args: self.on_entry_out_folder())

        self.name_pattern3 = Tk.StringVar()
        entry = Tk.Entry(iframe, textvariable=self.name_pattern3, width=20)
        entry.grid(row=grid_row, column=2, padx=10, sticky=Tk.W)

        self.num_files3 = Tk.IntVar()
        num_files_label = Tk.Label(iframe, textvariable=self.num_files3, width=num_files_label_width, justify=Tk.RIGHT)
        num_files_label.grid(row=grid_row, column=3)

        grid_row += 1
        space = Tk.Label(iframe)
        space.grid(row=grid_row, column=0)

        grid_row += 1
        label = Tk.Label(iframe, text="Progress")
        label.grid(row=grid_row, column=0)
        self.mpb = ttk.Progressbar(iframe,orient ="horizontal", length=600, mode="determinate")
        self.mpb.grid(row=grid_row, column=1, columnspan=3, padx=5)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        self.subtract_btn = Tk.Button(box, text="Subtract", width=10, command=self.subtract, state=Tk.DISABLED)
        self.subtract_btn.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def show(self):
        self._show()

    def on_entry_in_folder(self, in_folder, in_files, name_pattern, num_files):
        in_folder_ = in_folder.get()
        if is_empty_val(in_folder_) or in_folder_ == self.in_folder_prev:
            return

        self.in_folder_prev = in_folder_
        files = get_xray_files(in_folder_)
        in_files.clear()
        in_files += files
        if len(files) > 0:
            dir_, file = os.path.split(files[0])
            name_pattern.set(file)
        num_files.set( len(files) )
        if self.num_files1.get() > 0:
            self.maker_btn.config(state=Tk.NORMAL)

        filename = get_mtd_filename( in_folder_ )
        if filename is None:
            if self.use_mtd:
                self.in_folder1_entry.config(fg='red')
                self.update()
                MessageBox.showerror("Error", "No MTD file is found.", parent=self)
                return

        self.in_folder1_entry.config(fg='black')
        self.mtd_file_path = filename

        self.in_folder2.set(in_folder_ + '/bg')
        if len(files) > 0:
            name_pattern_re = re.compile(r'(\d+)(\.dat)$')
            bg_file = re.sub(name_pattern_re, lambda m: 'bg' + m.group(2), file)
            self.name_pattern2.set(bg_file)

        self.out_folder.set(in_folder_ + '/sub')
        if len(files) > 0:
            self.on_entry_out_folder()

        self.maker_btn_frame.start()

    def on_entry_out_folder(self):
        _, in_folder1 = os.path.split(self.in_folder1.get())
        name_pattern1 = self.name_pattern1.get()
        _, in_folder2 = os.path.split(self.in_folder2.get())
        name_pattern2 = self.name_pattern2.get()
        # print((in_folder1, name_pattern1), (in_folder2, name_pattern2))
        _, out_folder = os.path.split(self.out_folder.get())

        if len(name_pattern1) > 0:
            ret_pattern = self.generate_pattern(in_folder1, name_pattern1, in_folder2, name_pattern2, out_folder)
            self.name_pattern3.set(ret_pattern)

    def subtract(self):

        name_pattern = self.name_pattern3.get().replace('_00000', '_%05d')
        print('name_pattern=', name_pattern)

        self.busy = True
        background_data = self.background_info[0]
        process_option = 1

        out_folder = self.out_folder.get()
        if not os.path.exists(out_folder):
            mkdirs_with_retry(out_folder)

        subtract_impl(self.in_files1, background_data, process_option, out_folder, name_pattern, dialog=self)
        self.busy = False

    def is_busy(self):
        return self.busy

    def generate_pattern(self, in_folder1, name_pattern1, in_folder2, name_pattern2, out_folder):
        name_pattern_re = re.compile(r'(\d+)(\.dat)$')
        sub_filename = re.sub(name_pattern_re, lambda m: m.group(1) + '_sub' + m.group(2), name_pattern1)

        return sub_filename

    def show_backgroud_maker_dialog(self):
        from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
        from BackgroundMaker import BackgroundMakerDialog
        from molass_legacy._MOLASS.SerialSettings import set_setting, do_microfluidic_temporary_settings
        from SerialTestUtils import prepare_serialdata_env

        self.maker_btn_frame.stop()
        in_folder = self.in_folder1.get()

        if self.use_mtd:
            filename = self.mtd_file_path
            set_setting( 'use_xray_conc', 0 )
            set_setting( 'use_mtd_conc', 1 )
            set_setting( 'mtd_file_path', filename )
            do_microfluidic_temporary_settings()
            mtd_conc = True
        else:
            mtd_conc = False

        sd = prepare_serialdata_env( in_folder, root=self, mtd_conc=mtd_conc, logger=self.logger )
        pre_recog = PreliminaryRecognition(sd)

        dialog = BackgroundMakerDialog(self, pre_recog)
        dialog.show()
        if dialog.applied:
            self.background_info = dialog.get_the_background_info()
            self.save_background_data()
            self.num_files2.set(1)
            self.subtract_btn.config(state=Tk.NORMAL)

    def save_background_data(self):
        from molass_legacy.KekLib.NumpyUtils import np_savetxt_with_comments
        folder = self.in_folder2.get()
        if not os.path.exists(folder):
            mkdirs_with_retry(folder)

        filename = self.name_pattern2.get()
        filepath = os.path.join(folder, filename)
        background_data = self.background_info[0]
        elution_slice = self.background_info[1]
        comments = [
            '# averaged from %d to %d\n' % (elution_slice.start, elution_slice.stop-1)
            ]
        np_savetxt_with_comments(filepath, background_data, comments)



"""

    ファイル名：   AutoRunController.py

    処理内容：

        自動制御

    Copyright (c) 2020,2024, SAXS Team, KEK-PF
"""
import sys
import time
import logging

from molass_legacy.KekLib.OurTkinter     import Tk
import OurMessageBox  as MessageBox
from OurMock        import MagicMock

import ThreadsConnector
import ActionWindow
import threading

from SynthesizerSettings    import get_setting, get_mask
from Preferences            import get_preference
from ImageSynthesizer       import ImageSynthesizer
from DebugQueue             import debug_queue_get

import gettext
_ = gettext.gettext

class AutoRunController:

    def __init__( self, window, interval, image_info_table, log_file_path=None, on_stop=None ):
        self.window             = window
        self.interval           = interval
        self.image_info_table   = image_info_table
        self.log_file_path      = log_file_path
        self.synthesizer        = ImageSynthesizer()
        self.logger             = logging.getLogger( __name__ )
        self.on_stop            = on_stop

        print( '__init__:', threading.current_thread() )

        title       = _('Auto Run Thread Log')
        text        = _('Controlling syntheses' )
        geometry    = "900x400"
        self.action_window  = ActionWindow.ActionWindow(self.window, title, text, geometry, progressbarlabel=0, on_stop=self.on_stop )

    def start( self ):
        # テストでは self を GuiController.Controller から使用するため、
        # コンストラクタから分離しておく必要がある。
        conn = ThreadsConnector.ThreadsConnector( on_cancel=self.on_cancel )
        conn.runInGui( self.action_window, conn, None, self.control_syntheses, 'control' )

    def control_syntheses( self, connector, progress ):
        self.in_folder  = get_setting( 'in_folder' )
        self.adj_folder = get_setting( 'adj_folder' )
        self.syn_folder = get_setting( 'syn_folder' )
        # GUI から起動される場合、マスクは確定している。
        self.mask       = get_mask()

        print( 'control_syntheses:', threading.current_thread() )
        self.logger.info( 'Started auto-run with data end index %d.' % ( self.image_info_table.current_data_end) )

        while( True ):
            start_time  = time.time()
            progress.set( 0, self.interval )

            self.exec_thread = threading.Thread( None, self.exec_syntheses, 'exec', () )
            self.exec_thread.start()
            last_time   = time.time()
            # if self.exec_thread.is_alive():
            #    print( 'Looks like running.' )

            exec_thread_looks_like_done = False
            for i in range( self.interval ):
                # オーバヘッド分の遅れを調整する。
                sec_ = i + 1 - ( last_time - start_time )
                if sec_ < 0:
                    continue
                elif sec_ <= 1:
                    pass
                else:
                    sec_ = 1

                # print( '[%d] sec_=%g' % ( i, sec_ ) )

                time.sleep( sec_ )
                progress.tick()
                connector.ack()

                if not exec_thread_looks_like_done and not self.exec_thread.is_alive():
                    exec_thread_looks_like_done = True
                    # print( 'Looks like done.' )
                    # TODO: in case failed

                last_time   = time.time()

        self.logger.info(_('Unexpected! Never reach here.'))

    def exec_syntheses( self ):
        try:
            self.logger.info( 'Refreshing infomation.' )

            if debug_queue_get() == __name__ + '.die()': die()  # die(): undefined function

            selected_indices = self.image_info_table.get_not_yet_done_indices()
            num_selected_records = len( selected_indices )
            if num_selected_records > 0:
                self.logger.info( 'Begin sythesizing for %d samples.' % ( num_selected_records ) )

                # self.image_info_table.do_action( 3 )
                self.synthesizer.set_setting_info( 3 )

                exec_array = self.image_info_table.select_data_array( selected_indices )
                self.synthesizer.exec_array = exec_array

                mockconnector   = MagicMock()
                mockprogress    = MagicMock()

                self.synthesizer.exec_syntheses( mockconnector, mockprogress )

                self.image_info_table.update_current( selected_indices )
                self.image_info_table.clear_selection( selected_indices )
            else:
                self.logger.info( 'No new data since last refresh.' )
        except:
            logging.exception( 'Unexpected error in the execution thread:' )

        # self.image_info_table.refresh( log_file_path=self.log_file_path, bottom_view=True, logger=self.logger, autorun=True )
        self.window.refresh(autorun=True)

    def cancel( self ):
        self.action_window.cancel()

    def on_cancel( self ):
        # print( 'exec_cancel:', threading.current_thread() )

        if self.exec_thread.is_alive():
            self.exec_thread.join( 10 )

        self.exec_thread = None
        self.logger.info( 'exec_cancel' )



"""
    SMC.SolverLogLike.py

    Copyright (c) 2024, SAXS Team, KEK-PF    
"""
import numpy as np
import pytensor.tensor as pt
from pytensor.graph import Apply, Op

"""
    This implementation is based on the PyMC tutorial:
    https://www.pymc.io/projects/examples/en/latest/howto/blackbox_external_likelihood_numpy.html
    Using a “black box” likelihood function
    PyTensor Op without gradients
"""
class SolverLogLike(Op):
    def __init__(self, objective):
        self.objective = objective

    def make_node(self, params) -> Apply:
        # Convert inputs to tensor variables
        params = pt.as_tensor(params)

        inputs = [params]
        # Define output type, in our case a vector of likelihoods
        # with the same dimensions and same data type as data
        # If data must always be a vector, we could have hard-coded
        # outputs = [pt.vector()]
        outputs = [pt.dscalar()]

        # Apply is an object that combines inputs, outputs and an Op (self)
        return Apply(self, inputs, outputs)

    def perform(self, node: Apply, inputs: list[np.ndarray], outputs: list[list[None]]) -> None:
        # This is the method that compute numerical output
        # given numerical inputs. Everything here is numpy arrays
        params, = inputs  # this will contain my variables

        # call our numpy log-likelihood function
        loglike_eval = -self.objective(params)
        # self.logger.info("loglike_eval=%s", loglike_eval)

        # Save the result in the outputs list provided by PyTensor
        # There is one list per output, each containing another list
        # pre-populated with a `None` where the result should be saved.
        outputs[0][0] = np.asarray(loglike_eval)

"""
pytensor.compile.function.types.UnusedInputError:
pytensor.function was asked to create a function computing outputs given certain inputs,
but the provided input variable at index 0 is not part of the computational graph
needed to compute the outputs: joined_inputs.
To make this error into a warning,
you can pass the parameter on_unused_input='warn' to pytensor.function.
To disable it completely, use on_unused_input='ignore'.
"""


# coding: utf-8
"""
    GmmUtils.py.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from sklearn.mixture import GaussianMixture
from lmfit import minimize, Parameters
import molass_legacy.KekLib.DebugPlot as dplt
from Prob.ProbDataUtils import generate_sample_data, plot_hist_data

def get_gmm(size, data, k, max_iter=100):
    print('get_gmm: max_iter=', max_iter)
    gmm = GaussianMixture(k, max_iter=max_iter, covariance_type='spherical')
    gmm.fit(X=np.expand_dims(data,1))
    return gmm

def get_gmm__(size, data, k):
    from .SingleGmm import SingleGmm
    gmm = SingleGmm(k, max_iter=1000)
    gmm.fit(X=np.expand_dims(data,1))
    for k_ in range(k):
        print('gamma[0:5,%d]=' % k_, gmm.gamma[0:5,k_])
    return gmm

def get_gmm___(size, data, k):
    from .ZhiyaZuoGmm import GmmAdaptor
    gmm = GmmAdaptor(k, max_iter=20)
    gmm.fit(X=np.expand_dims(data,1))
    return gmm

def get_gmm____(size, data, k):
    from .MatsukenGmm import GMM
    gmm = GMM(k, max_iter=20)
    gmm.fit(X=np.expand_dims(data,1))
    return gmm

def gaussian(x, mu, sigma, scale):
    return scale * np.exp( - (x - mu)**2 / (2 * sigma**2) )

def get_gaussians(x, y, gmm):
    gy_list = []
    ty = np.zeros(len(x))
    for k in range(len(gmm.weights_)):
        w = gmm.weights_[k]
        m = gmm.means_[k]
        v = gmm.covariances_[k]
        print([k], w, m, v)
        gy = gaussian(x, m, np.sqrt(v), w)
        gy_list.append([gy, m])
        ty += gy

    sorted_gy_list = sorted(gy_list, key=lambda x:x[1])

    def obj_func(params):
        S   = params['S']
        return ty*S - y

    params = Parameters()
    S_init = np.max(y)/np.max(ty)
    params.add('S', value=S_init, min=0, max=S_init*100 )
    result = minimize(obj_func, params, args=())

    scale = result.params['S'].value
    return [ scale*rec[0] for rec in sorted_gy_list ], ty*scale

def plot_components(ax, x, cgys):
    for k, gy in enumerate(cgys):
        ax.plot(x, gy, ':', label='component-%d' % k)

def plot_gmm_elution(y1, y2, data1, data2, k, max_iter=100, plot_class=plt):
    from DataUtils import get_in_folder

    gmm1 = get_gmm(len(y1), data1, k, max_iter=max_iter)
    gmm2 = get_gmm(len(y2), data2, k, max_iter=max_iter)

    fig, axes = plot_class.subplots(nrows=1, ncols=2, figsize=(14,7))
    ax1, ax2 = axes

    fig.suptitle("Decomposition of %s using Gaussian Mixture Model" % get_in_folder(), fontsize=20)
    ax1.set_title("UV Elution Decomposition", fontsize=16)
    ax2.set_title("X-ray Elution Decomposition", fontsize=16)
    ax1.plot(y1, label='input data')
    ax2.plot(y2, label='data')

    x1 = np.arange(len(y1))
    cgys1, gy1 = get_gaussians(x1, y1, gmm1)
    ax1.plot(gy1, label='gmm-fit')
    plot_components(ax1, x1, cgys1)

    x2 = np.arange(len(y2))
    cgys2, gy2 = get_gaussians(x2, y2, gmm2)
    ax2.plot(gy2, label='gmm-fit')
    plot_components(ax2, x2, cgys2)

    ax1.legend()
    ax2.legend()

    fig.tight_layout()
    fig.subplots_adjust(top=0.88)

def spike_demo(in_folder, lpm_correct=False, n_components=2, max_iter=100):
    from RawData import RawXray, RawUv
    ru = RawUv(in_folder)
    i = ru.get_row_index(280)
    y1 = ru.data[i,:]
    if lpm_correct:
        from LPM import get_corrected
        y1 = get_corrected(y1)

    rx = RawXray(in_folder)
    i = rx.get_row_index(0.02)
    y2 = rx.data[i,:]
    if lpm_correct:
        from LPM import get_corrected
        y2 = get_corrected(y2)

    data1 = generate_sample_data(y1, 2)
    data2 = generate_sample_data(y2, 2)

    dplt.push()
    plot_hist_data(y1, y2, data1, data2, plot_class=dplt)
    dplt.show()
    dplt.pop()

    dplt.push()
    plot_gmm_elution(y1, y2, data1, data2, n_components, max_iter=max_iter, plot_class=dplt)
    dplt.show()
    dplt.pop()



"""
    Wave.Circle.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

class BeamAnim:
    def __init__(self, ):
        pass


def circle_xy(cx, cy, r, n=100):
    t = np.linspace(0, np.pi*2, n)
    return cx + r*np.cos(t), cy + r*np.sin(t)

def demo1():
    fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(21, 7))

    ax1.plot(0, 0, 'o')

    z = np.exp(np.pi/4*1j)
    print(z)

    ax1.plot(z.real, z.imag, 'o')
    x, y = circle_xy(0, 0, 1)
    ax1.plot(x, y)

    fig.tight_layout()
    plt.show()



"""
    GuinierRg.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier

def compute_rg_params_impl(lrf_info):
    print("compute_rg_params_impl")
    qv = lrf_info.qv
    P = lrf_info.matrices[0]
    P_ = P[:,:-1]           # excluding baseline
    D = lrf_info.xrD
    E = lrf_info.xrE
    print(P.shape, D.shape, E.shape)
    """
        M = PC
        P = MC⁺
        Pe = sqrt(E**2 @ W**2)
    """
    D_pinv = np.linalg.pinv(D)
    W   = np.dot(D_pinv, P_)
    Pe  = np.sqrt(np.dot(E**2, W**2))

    ret_rgs = []
    qualities = []
    for y, ye in zip(P_.T, Pe.T):
        sg = SimpleGuinier(np.array([qv, y, ye]).T)
        ret_rgs.append(sg.Rg)
        qualities.append(sg.basic_quality)

    return np.array(ret_rgs), np.array(qualities)



# coding: utf-8
"""
    GuinierInterval.py

    Copyright (c) 2016-2017, Masatsuyo Takahashi, KEK-PF
"""
import sys
import numpy                as np
from bisect                 import bisect_right
from scipy.interpolate      import UnivariateSpline
from lmfit                  import minimize, Parameters, Parameter, report_fit
from SmoothSpline           import SmoothSpline

DEBUG = False

WeightsArray = [
    np.array( [ 100, 10,  1,  1 ] ),
    ]

MIN_SIZE = 1e-6

class InvervalEvaluationModel:
    def __init__( self, Rg, degree, x, smoother, weights, qmax_limit ):
        self.Rg     = Rg
        self.degree = degree
        self.x      = x
        self.dx     = x[1] - x[0]
        # self.qrg_q_limit    = np.sqrt( 3 * degree / 2 )
        self.qmax_limit = qmax_limit
        if DEBUG: print( 'InvervalEvaluationModel: qmax_limit=', qmax_limit )
        self.smoother = smoother
        self.weights = weights

    def __call__( self, params ):
        min_q   = params[ 'min_q' ].value
        max_q   = params[ 'max_q' ].value
        # max_q   = self.qmax_limit

        size_ = max_q - min_q

        Rg_array = []
        Rg_diff_array = []
        min_qq = min_q**2

        if size_ > MIN_SIZE:
            min_2   = min_q + self.dx*2     # *2 seems to make it a little better
            min_22  = min_2**2

            if self.smoother.is_guinier:
                y1  = self.smoother( min_qq )
                y2  = self.smoother( min_22 )
            else:
                y1 = np.log( self.smoother( min_q ) )
                y2  = np.log( self.smoother( min_2 ) )
            a1   = ( y2 - y1 ) / ( min_22 - min_qq )
            a_array = [a1]

            xmax_ = min( max_q, self.qmax_limit )
            if xmax_ > min_2:
                max_qq = xmax_**2
                if self.smoother.is_guinier:
                    y3  = self.smoother( max_qq )
                else:
                    y3 = np.log( self.smoother( max_q ) )
                a2  = ( y3 - y2 ) / ( max_qq - min_22 )
                a_array.append( a2 )
                # print( 'a_array=', a_array )
            else:
                a_array.append( a1 )

            for a in a_array:
                if a < 0:
                    Rg = np.sqrt( -a*3 )
                    Rg_diff = abs( Rg - self.Rg )**2
                    """
                    ex = max_q * Rg - self.qrg_q_limit
                    if ex > 0:
                        Rg_diff += ex
                    """
                else:
                    # raise Exception( 'a=%g >= 0' % a )
                    Rg = None
                    Rg_diff = 9999 + size_**2

                Rg_array.append( Rg )
                Rg_diff_array.append( Rg_diff )
        else:
            Rg = np.nan
            Rg_diff = 999 + size_**2
            Rg_array = [ Rg, Rg ]
            Rg_diff_array = [ Rg_diff, Rg_diff ]

        size_factor = 1e-2/size_**2 if size_ > MIN_SIZE else 1e-2/MIN_SIZE**2 + size_**2

        # start_factor = min_q ** 2 / max( 0.01, Rg_diff ) / size_factor * 1e5
        # start_factor = min_q / size_factor * 10
        # start_factor = min_q ** 2 / size_factor * 10000
        # start_factor = min_q ** 4 * size_factor * 1e5
        # start_factor = min_q ** 2 * size_factor**2 * 1e3
        # start_factor = min_q**2 * np.sqrt( size_factor ) * 10000
        # start_factor = min_q ** 4 * 10000
        start_factor = min_qq * 1e4

        # Rg_diff_factor = np.sqrt( np.sum( ( self.weights[0:2] * Rg_diff_array)**2 ) )
        # ret = np.array( [ Rg_diff_factor, self.weights[2]*size_factor, self.weights[3]*start_factor ] )
        ret = self.weights * np.array( [ Rg_diff_array[0], Rg_diff_array[1], size_factor, start_factor ] )

        if DEBUG:
            Rg_str = 'None' if Rg is None else '%0.3g' % Rg
            """
            print( '%.2g' % self.degree, '(%0.3g, %0.3g)' % (min_q, max_q),
                    ', Rg=%s' % Rg_str, ', Rg_diff=(%0.3g, %0.3g)' % tuple( Rg_diff_array ) )
            """
            opt_v = np.inner( ret, ret )
            # print( 'min_q=%.4g => (%0.3g, %0.3g) ret=[ %.3g %.3g %.3g %0.3g ], %.4g' % ( ( min_q, ) + tuple( Rg_diff_array ) + tuple( ret ) + ( opt_v, ) ) )
            print( 'Rg=%.4g, min_q=%.4g => (%.3g, %.3g), %.4g' % ( ( self.Rg, min_q, ) + tuple( Rg_array ) + ( opt_v, ) ) )

        return ret

class GuinierInterval:
    def __init__( self, Q, I, x, y, w, fit, qrg_limits ):
        self.Q      = Q
        self.I      = I
        self.x      = x
        self.y      = y
        self.w      = w
        self.fit    = fit
        self.Rg     = fit.Rg
        self.degree = fit.degree
        self.qmax_limit = min( qrg_limits[1] / fit.Rg, x[-1] )
        self.qmax_index = bisect_right( self.Q, self.qmax_limit )
        try:
            # self.spline = UnivariateSpline( x, y, k=3, s=5e8 )
            # be aware that x and y are non-transformed values,
            # i.e. not x**2 nor log(y)
            # self.spline = UnivariateSpline( x, y, w=w, k=3, s=1e8 )
            # smooth_spline = SmoothSpline( x, y, w=w, sf_log_base=30 )
            self.smoother = SmoothSpline( x, y, w=w, is_guinier=False )
            # self.smoother = SmootherChain( x, y, is_guinier=False )
            # self.smoother = fit.smoother
        except Exception as e:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            if True:
                etb = ExceptionTracebacker()
                print( etb )
                print( 'y=', y )
            # TODO: raise Exception
            self.smoother = None

    def get_optimal_interval( self ):

        # print( 'qmax_limit=', self.qmax_limit )
        min_q_inits = list( map( lambda p: (1 - p) * self.x[0] + p * self.qmax_limit, [ 0.1, 0.2, 0.5 ] ) )
        # print( 'min_q_inits=', min_q_inits )

        for weights in WeightsArray:

            opt_redchi = None
            opt_result = None
            opt_q_init = None

            for min_q_init in min_q_inits:
                if DEBUG: print( 'minimize with min_q_init', min_q_init )

                model   = InvervalEvaluationModel( self.Rg, self.degree, self.x, self.smoother, weights, self.qmax_limit )

                params  = Parameters()

                params.add('min_q',    value= min_q_init,       min=self.x[0], max=self.qmax_limit )
                params.add('max_q',    value= self.qmax_limit,  min=self.x[0], max=self.qmax_limit )

                result = minimize( model, params, args=() )

                if opt_redchi is None or result.redchi < opt_redchi:
                    opt_redchi = result.redchi
                    opt_result = result
                    opt_q_init = min_q_init

            self.min_q = opt_result.params[ 'min_q' ].value

            if DEBUG:
                print( 'minimized with opt_q_init=%.4g' % opt_q_init )
                print( 'optimal min_q=%.4g' % self.min_q )

            self.max_q = result.params[ 'max_q' ].value
            # self.max_q = self.qmax_limit

            if DEBUG: print( self.min_q, self.max_q )            

            f_ = bisect_right( self.Q, self.min_q )
            t_ = bisect_right( self.Q, self.max_q )
            if DEBUG: print( 'get_optimal_interval', self.min_q, self.max_q, f_, t_, self.qmax_index )
            # sys.exit()

            # TODO: better check
            if f_ < self.qmax_index:
                if DEBUG: print( self.min_q, self.max_q, f_, t_ )
                return self.min_q, self.max_q, f_, t_

        raise Exception( 'Guninier interval optimazation failed!' )



# coding: utf-8
"""
    UvBaseSurface.py

    Copyright (c) 2017-2020, SAXS Team, KEK-PF
"""
import numpy        as np
from molass_legacy.Test.TesterLogger import write_to_tester_log

USE_CONSISTENT_BASE         = True
MIN_USABLE_SIZE_IN_SIGMA    = 3
PENALTY_RATE                = 100
ACCEPTABLE_OFFSSET          = 0.05  #  0.053 for Dec05, 0.07 for SUB_TRN1
SAFE_ENOUGH_POS_SIGMA       = 5

class UvBaseSurface:
    def __init__( self, rail_points, tailcurve, rails, rail_points_sigma, max_y,
                    estimeted_rails_base,
                    rail_points_offset=None ):
        assert len(rails) == 2, 'len(rails) != 2'

        # print( 'slope=', slope )
        self.tailcurve  = tailcurve
        self.rails      = rails

        """
            y = A*x + B
        """
        self.jsize  = rail_points[1] - rail_points[0]
        j   = np.arange( self.jsize + 1 )

        if USE_CONSISTENT_BASE:
            B0  = estimeted_rails_base[0] - tailcurve[0]
            B1  = estimeted_rails_base[1] - tailcurve[-1]
        else:

            rail_points_ok_pre  = [ abs( rail_points_sigma[i] ) >= MIN_USABLE_SIZE_IN_SIGMA  for i in range(2) ]
            if rail_points_offset is None:
                rail_points_offset = [ abs( rail[0] - estimeted_rails_base[i] )/max_y for i, rail in enumerate(rails) ]

            print( 'rail_points_offset=', rail_points_offset )
            rail_points_ok  = [ abs( rail_points_sigma[i] ) - max( 0, rail_points_offset[i] - ACCEPTABLE_OFFSSET )* PENALTY_RATE  >= SAFE_ENOUGH_POS_SIGMA
                                or (  b and rail_points_offset[i] <= ACCEPTABLE_OFFSSET ) for i, b in enumerate(rail_points_ok_pre) ]
            write_to_tester_log( 'rail_points_ok=' + str(rail_points_ok) + '\n' )
            print( 'rail_points_ok=', rail_points_ok )
            print( 'estimeted_rails_base=', estimeted_rails_base )

            if rail_points_ok[0] and rail_points_ok[1]:
                B0  = rails[0][0] - rails[0][-1]
                B1  = rails[1][0] - rails[1][-1]
            elif rail_points_ok[0]:
                B0  = rails[0][0] - rails[0][-1]
                B1  = estimeted_rails_base[1] - rails[1][-1]
            elif rail_points_ok[1]:
                B0  = estimeted_rails_base[0] - rails[0][-1]
                B1  = rails[1][0] - rails[1][-1]
            else:
                B0  = estimeted_rails_base[0] - rails[0][-1]
                B1  = estimeted_rails_base[1] - rails[1][-1]

        intercept = B0
        slope = ( B1 - B0 ) / self.jsize
        self.baseshift  = slope*j + intercept

        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            fig = plt.figure()
            ax = fig.gca()
            ax.plot(self.baseshift)
            ax.plot(self.tailcurve)
            ax.plot(self.tailcurve + self.baseshift)
            fig.tight_layout()
            plt.show()

    def get_basecurve( self ):
        basecurve   = self.tailcurve + self.baseshift
        return basecurve

    def debug_plot( self ):
        pass



"""
    IntensitySmootherAveraging.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""
import numpy as np
from molass_legacy.SerialAnalyzer.ElutionalSmootherUtils import LightIntensityData

class IntensitySmootherAveraging:
    def __init__( self, intensity_array, num_curves_averaged ):
        # print( 'IntensitySmootherAveraging: len(intensity_array)=', len(intensity_array),
        #        ', num_curves_averaged=', num_curves_averaged )
        if type( intensity_array ) == np.ndarray:
            data_array = intensity_array
        else:
            data_array = [ intensity.orig_array  for intensity in intensity_array ]
        self.intensity_matrix   = np.array( data_array )
        self.offset_start       = - ( num_curves_averaged//2 )
        self.offset_stop        = self.offset_start + num_curves_averaged
        # print( 'self.intensity_matrix.shape=', self.intensity_matrix.shape )

    def __call__( self, indeces, return_numpy_ndarray=False ):

        intensity_array = []
        slice_array = []
        for i in indeces:
            start   = max( 0, i + self.offset_start )
            stop    = min( self.intensity_matrix.shape[0], i + self.offset_stop )
            slice_  = slice(start, stop)
            data_qi = np.average( self.intensity_matrix[ slice_, :, 0:2 ], axis=0 )
            # apply error propagation rule to error data
            error2  = self.intensity_matrix[ slice_, :, 2 ]**2
            data_e_ = np.sqrt( np.sum( error2, axis=0 ) ).reshape( ( error2.shape[1], 1) )/( stop - start )
            data = np.hstack( [ data_qi, data_e_ ] )
            if return_numpy_ndarray:
                intensity_array.append( data )
                slice_array.append( slice_ )
            else:
                smoothed_intensity = LightIntensityData( orig_array=data )
                intensity_array.append( smoothed_intensity )

        if return_numpy_ndarray:
            return np.array( intensity_array ), slice_array
        else:
            return intensity_array

class ConcentrationSmootherAveraging:
    def __init__( self, c_vector, num_curves_averaged ):
        self.c_vector       = c_vector
        self.offset_start   = - ( num_curves_averaged//2 )
        self.offset_stop    = self.offset_start + num_curves_averaged

    def __call__( self, indeces ):
        concentration_array = []
        for i in indeces:
            start   = max( 0, i + self.offset_start )
            stop    = min( len( self.c_vector ), i + self.offset_stop )
            concentration_array.append( np.average( self.c_vector[start:stop] ) )

        return np.array( concentration_array )
